"""
@author:    José Miguel Algarín
@email:     josalggui@i3m.upv.es
@affiliation: MRILab, i3M, CSIC, Valencia, Spain
"""

import datetime
from PyQt5.QtWidgets import (
    QMainWindow, QStatusBar, QWidget, QHBoxLayout, QVBoxLayout, QTableWidget,
    QSizePolicy
)
from PyQt5.QtCore import QThreadPool
import qdarkstyle

from marge.controller.controller_console import ConsoleController
from marge.controller.controller_figures import FiguresLayoutController
from marge.controller.controller_history_list import HistoryListController
from marge.controller.controller_menu import MenuController
from marge.controller.controller_protocol_inputs import ProtocolInputsController
from marge.controller.controller_protocol_list import ProtocolListController
from marge.controller.controller_toolbar_figures import FiguresController
from marge.controller.controller_toolbar_marcos import MarcosController
from marge.controller.controller_toolbar_protocols import ProtocolsController
from marge.controller.controller_toolbar_sequences import SequenceController
from marge.controller.controller_sequence_list import SequenceListController
from marge.controller.controller_sequence_inputs import SequenceInputsController
from marge.widgets.widget_custom_and_protocol import CustomAndProtocolWidget
from marge.controller.controller_postprocessing import ProcessingWindowController


class MainWindow(QMainWindow):
    def __init__(self, session, demo=False, parent=None):
        super(MainWindow, self).__init__()
        self.app_open = True
        self.toolbar_sequences = None
        self.toolbar_marcos = None
        self.session = session
        self.demo = demo
        self.parent = parent
        self.setWindowTitle(session.get('directory', 'MaRGE'))
        self.setGeometry(20, 40, 1680, 720)

        # Threadpool for parallel running
        self.threadpool = QThreadPool()

        # Set stylesheet based on theme
        if self.session["black_theme"]:
            self.styleSheet = qdarkstyle.load_stylesheet_pyqt5()
        else:
            self.styleSheet = ""
        self.setStyleSheet(self.styleSheet)

        # Create console
        now = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        log_name = f"log_{now}.txt"
        self.console = ConsoleController(log_name)

        # Add marcos toolbar
        self.toolbar_marcos = MarcosController(self, "MaRCoS toolbar")
        self.addToolBar(self.toolbar_marcos)

        # Add sequence toolbar
        self.toolbar_sequences = SequenceController(self, "Sequence toolbar")
        self.addToolBar(self.toolbar_sequences)

        # Add image toolbar
        self.toolbar_figures = FiguresController(self, "Figures toolbar")
        self.addToolBar(self.toolbar_figures)

        # Add protocol toolbar
        self.toolbar_protocols = ProtocolsController(self, "Protocols toolbar")
        self.addToolBar(self.toolbar_protocols)

        # Add Scanner menu
        self.menu = self.menuBar()
        MenuController(main=self)

        # Status bar
        self.setStatusBar(QStatusBar(self))

        # Create central widget that will contain the layout
        self.main_widget = QWidget()
        self.setCentralWidget(self.main_widget)

        # Add main layout to input other layouts
        self.layout_main = QHBoxLayout()
        self.main_widget.setLayout(self.layout_main)

        # Layout for sequences and console
        self.layout_left = QVBoxLayout()
        self.layout_main.addLayout(self.layout_left)

        # Layout for outputs
        self.layout_right = QVBoxLayout()
        self.layout_main.addLayout(self.layout_right)

        # Add custom_and_protocol widget
        self.custom_and_protocol = CustomAndProtocolWidget()
        self.layout_left.addWidget(self.custom_and_protocol)

        # Add sequence list to custom tab
        self.sequence_list = SequenceListController(parent=self)
        self.custom_and_protocol.custom_layout.addWidget(self.sequence_list)

        # Add sequence inputs to custom tab
        self.sequence_inputs = SequenceInputsController(parent=self)
        self.custom_and_protocol.custom_layout.addWidget(self.sequence_inputs)

        # Add protocols list to protocol tab
        self.protocol_list = ProtocolListController(main=self)
        self.custom_and_protocol.protocol_layout.addWidget(self.protocol_list)

        # Add protocol sequences to protocol tab
        self.protocol_inputs = ProtocolInputsController(main=self)
        self.custom_and_protocol.protocol_layout.addWidget(self.protocol_inputs)

        # Add console
        self.layout_left.addWidget(self.console)

        # Add layout to show the figures
        self.figures_layout = FiguresLayoutController(self)
        self.figures_layout.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.layout_right.addWidget(self.figures_layout)

        # Layout for output history
        self.layout_right_h = QHBoxLayout()
        self.layout_right.addLayout(self.layout_right_h)

        # Add list to show the history
        self.history_list = HistoryListController(parent=self)
        self.layout_right_h.addWidget(self.history_list)
        self.history_list.setMaximumHeight(200)
        self.history_list.setMinimumHeight(200)

        # Table with input parameters from historic images
        self.input_table = QTableWidget()
        self.input_table.setMaximumHeight(200)
        self.input_table.setMinimumHeight(200)
        self.layout_right_h.addWidget(self.input_table)

        # Create the post-processing toolbox
        self.post_gui = ProcessingWindowController(session=self.session, main=self)




#  NanoVNASaver
#
#  A python program to view and export Touchstone data from a NanoVNA
#  Copyright (C) 2019, 2020  Rune B. Broberg
#  Copyright (C) 2020,2021 NanoVNA-Saver Authors
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
import struct
from typing import List

import serial
import numpy as np
from PyQt5 import QtGui

from marge.vna.Serial import drain_serial, Interface
from marge.vna.VNA import VNA

logger = logging.getLogger(__name__)


class TinySA(VNA):
    name = "tinySA"
    screenwidth = 320
    screenheight = 240
    valid_datapoints = (290, )

    def __init__(self, iface: Interface):
        super().__init__(iface)
        self.features = {'Screenshots'}
        logger.debug("Setting initial start,stop")
        self.start, self.stop = self._get_running_frequencies()
        self.sweep_max_freq_Hz = 950e6
        self._sweepdata = []

    def _get_running_frequencies(self):

        logger.debug("Reading values: frequencies")
        try:
            frequencies = super().readValues("frequencies")
            return frequencies[0], frequencies[-1]
        except Exception as e:
            logger.warning("%s reading frequencies", e)
            logger.info("falling back to generic")

        return VNA._get_running_frequencies(self)

    def _capture_data(self) -> bytes:
        timeout = self.serial.timeout
        with self.serial.lock:
            drain_serial(self.serial)
            self.serial.write("capture\r".encode('ascii'))
            self.serial.readline()
            self.serial.timeout = 4
            image_data = self.serial.read(
                self.screenwidth * self.screenheight * 2)
            self.serial.timeout = timeout
        self.serial.timeout = timeout
        return image_data

    def _convert_data(self, image_data: bytes) -> bytes:
        rgb_data = struct.unpack(
            f">{self.screenwidth * self.screenheight}H",
            image_data)
        rgb_array = np.array(rgb_data, dtype=np.uint32)
        return (0xFF000000 +
                ((rgb_array & 0xF800) << 8) +
                ((rgb_array & 0x07E0) << 5) +
                ((rgb_array & 0x001F) << 3))

    def getScreenshot(self) -> QtGui.QPixmap:
        logger.debug("Capturing screenshot...")
        if not self.connected():
            return QtGui.QPixmap()
        try:
            rgba_array = self._convert_data(self._capture_data())
            image = QtGui.QImage(
                rgba_array,
                self.screenwidth,
                self.screenheight,
                QtGui.QImage.Format_ARGB32)
            logger.debug("Captured screenshot")
            return QtGui.QPixmap(image)
        except serial.SerialException as exc:
            logger.exception(
                "Exception while capturing screenshot: %s", exc)
        return QtGui.QPixmap()

    def resetSweep(self, start: int, stop: int):
        return

    def setSweep(self, start, stop):
        self.start = start
        self.stop = stop
        list(self.exec_command(f"sweep {start} {stop} {self.datapoints}"))
        list(self.exec_command("trigger auto"))

    def readFrequencies(self) -> List[int]:
        logger.debug("readFrequencies")
        return [int(line) for line in self.exec_command("frequencies")]

    def readValues(self, value) -> List[str]:
        logger.debug("Read: %s", value)
        if value == "data 0":
            self._sweepdata = [f"0 {line.strip()}"
                               for line in self.exec_command("data")]
        return self._sweepdata



import sys
import os
import csv
import numpy as np
from marge.configs import hw_config as hw
from PyQt5.QtWidgets import (
    QApplication, QWidget, QVBoxLayout, QLineEdit, QPushButton,
    QHBoxLayout, QLabel, QGridLayout
)

class GradientsWidget(QWidget):
    def __init__(self):
        super().__init__()

        # Main layout
        self.main_layout = QVBoxLayout()
        self.layout = QGridLayout()
        self.main_layout.addLayout(self.layout)

        # Parameters to save inputs
        self.labels = []
        self.values = []
        self.tips = []

        # Add inputs
        self.add_input(label="Gx max (mT/m)", value="50", tip="Maximum gradient along X direction")
        self.add_input(label="Gy max (mT/m)", value="80", tip="Maximum gradient along Y direction")
        self.add_input(label="Gz max (mT/m)", value="70", tip="Maximum gradient along Z direction")
        self.add_input(label="Max slew rate (mT/m/ms)", value="80", tip="Maximum slew rate")
        self.add_input(label="Gradient raster time (us)", value="50", tip="Time resolution of gradient system")
        self.add_input(label="Gradient rise time (us)", value="400", tip="Time to reach maximum gradient amplitude")
        self.add_input(label="Gradient steps", value="16", tip="Number of steps for gradient transition")
        self.add_input(label="Gradient delay (us)", value="9", tip="Delay before gradient application")
        self.add_input(label="GPA model", value="None",
                       tip="GPA model: Only 'Barthel' is supported; otherwise, leave it empty.")

        # Dictionary to store references to input fields
        self.input_boxes = {}

        # Create blocks iteratively
        for row, (label, value) in enumerate(zip(self.labels, self.values)):
            label_widget = QLabel(label)
            input_box = QLineEdit(value)
            input_box.setStatusTip(self.tips[row])
            self.input_boxes[label] = input_box

            self.layout.addWidget(label_widget, row, 0)  # Label in column 0
            self.layout.addWidget(input_box, row, 1)     # Input box in column 1

        # Buttons
        self.save_button = QPushButton('Save', self)
        self.save_button.clicked.connect(self.save_gradient_entries)

        layout = QHBoxLayout()
        layout.addWidget(self.save_button)
        self.main_layout.addLayout(layout)
        self.main_layout.addStretch()

        self.setLayout(self.main_layout)
        self.setWindowTitle('Gradients Entry')
        self.resize(400, 200)

        # Load saved gradient entries
        self.load_gradient_entries()

        # Update hardware
        self.update_hw_config_gradients()

    def add_input(self, label="", value="", tip=""):
        self.labels.append(label)
        self.values.append(value)
        self.tips.append(tip)

    def update_hw_config_gradients(self):
        hw.gFactor = np.array([float(self.input_boxes["Gx max (mT/m)"].text()),
                               float(self.input_boxes["Gy max (mT/m)"].text()),
                               float(self.input_boxes["Gz max (mT/m)"].text())]) * 1e-3  # T/m
        hw.max_grad = np.min(hw.gFactor) * 1e3
        hw.max_slew_rate = float(self.input_boxes["Max slew rate (mT/m/ms)"].text())
        hw.grad_raster_time = float(self.input_boxes["Gradient raster time (us)"].text()) * 1e-6  # s
        hw.grad_rise_time = float(self.input_boxes["Gradient rise time (us)"].text()) * 1e-6  # s
        hw.grad_steps = int(self.input_boxes["Gradient steps"].text())
        hw.gradDelay = int(self.input_boxes["Gradient delay (us)"].text())
        hw.gpa_model = self.input_boxes["GPA model"].text()

    def save_gradient_entries(self):
        file_name = "configs/hw_gradients.csv"

        # ✅ Ensure the 'configs' directory exists
        os.makedirs(os.path.dirname(file_name), exist_ok=True)

        with open(file_name, 'w', newline='') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(["ID", "Value"])
            for label, input_box in self.input_boxes.items():
                writer.writerow([label, input_box.text()])
        self.update_hw_config_gradients()
        print(f"Data saved for gradients.")

    def load_gradient_entries(self):
        """Load label-value pairs from a CSV file and update the input fields."""
        file_path = "configs/hw_gradients.csv"
        try:
            with open(file_path, mode="r", newline="") as file:
                reader = csv.reader(file)
                next(reader)  # Skip header row
                for row in reader:
                    if len(row) == 2:
                        label, value = row
                        if label in self.input_boxes:
                            self.input_boxes[label].setText(value)
        except:
            print("No hardware configuration loaded for gradients.")

if __name__ == '__main__':
    app = QApplication(sys.argv)
    widget = GradientsWidget()
    widget.show()
    sys.exit(app.exec())



## IP address: RP address or 'localhost' if emulating a local server.
## Uncomment one of the lines below.
#ip_address = "localhost"
ip_address = "192.168.1.101"

## Port: always 11111 for now
port = 11111

## FPGA clock frequency: uncomment one of the below to configure various
## system behaviour. Right now only 122.88 is supported.
fpga_clk_freq_MHz = 122.88 # RP-122
#fpga_clk_freq_MHz = 125.0 # RP-125

## Gradient board: uncomment one of the below to configure the gradient data format
grad_board = "gpa-fhdo"
#grad_board = "ocra1"

## GPA-FHDO current per volt setting (determined by resistors)
gpa_fhdo_current_per_volt = 2.5

## Flocra-pulseq path, for use of the flocra-pulseq library (optional).
## Uncomment the lines below and adjust the path to suit your
## flocra-pulseq location.
#import sys
#sys.path.append('/home/vlad/Documents/mri/flocra-pulseq')



import os
import time
import threading
import numpy as np
from marge.widgets.widget_reconstruction import ReconstructionTabWidget
from marge.marge_utils import utils

try:
    import cupy as cp
except ImportError:
    pass


def getPath():
    """
    Get the absolute path to the MATLAB script file.

    Returns:
        str: Absolute path to the MATLAB script file.
    """
    # Get the absolute path of the current Python script directory
    current_dir = os.path.dirname(os.path.abspath(__file__))

    # Relative path to the "scripts" directory in your project
    scripts_dir = os.path.join(current_dir, '..', 'scripts')

    # Absolute path to the "art.m" file
    matlab_script_path = os.path.join(scripts_dir, 'art.m')

    return matlab_script_path


def ramp(kSpace, n, m, nb_point):
    """
    Apply a ramp filter to the k-space data.

    Args:
        kSpace (ndarray): K-space data.
        n (int): Number of zero-filled points in k-space.
        m (int): Number of acquired points in k-space.
        nb_point (int): Number of points before m+n where reconstruction begins to go to zero.

    Returns:
        ndarray: K-space data with the ramp filter applied.
    """
    kSpace_ramp = np.copy(kSpace)
    kSpace_ramp[:, :, n + m::] = 0.0

    # Index of the gradient
    start_point = n + m - nb_point
    end_point = n + m
    gradient_factor = kSpace_ramp[:, :, n + m - nb_point] / nb_point

    # Go progressively to 0
    for i in range(start_point, end_point):
        kSpace_ramp[:, :, i + 1] *= kSpace_ramp[:, :, i] - gradient_factor

    return kSpace_ramp


def hanningFilter(kSpace, mm, nb_point):
    """
    Apply a Hanning filter to the k-space data.

    Args:
        kSpace (ndarray): K-space data.
        n (int): Number of zero-filled points in k-space.
        m (int): Number of acquired points in k-space.
        nb_point (int): Number of points before m+n where reconstruction begins to go to zero.

    Returns:
        ndarray: K-space data with the Hanning filter applied.
    """
    kSpace_hanning = np.copy(kSpace)
    kSpace_hanning[mm[0]::, :, :] = 0.0
    kSpace_hanning[:, mm[1]::, :] = 0.0
    kSpace_hanning[:, :, mm[2]::] = 0.0

    # Calculate the Hanning window
    hanning_window = np.hanning(nb_point * 2)
    hanning_window = hanning_window[int(len(hanning_window)/2)::]

    if not mm[0] == np.size(kSpace, 0):
        for ii in range(nb_point):
            kSpace_hanning[mm[0]-nb_point+ii+1, :, :] *= hanning_window[ii]
    if not mm[1] == np.size(kSpace, 1):
        for ii in range(nb_point):
            kSpace_hanning[:, mm[1]-nb_point+ii+1, :] *= hanning_window[ii]
    if not mm[2] == np.size(kSpace, 2):
        for ii in range(nb_point):
            kSpace_hanning[:, :, mm[2]-nb_point+ii+1] *= hanning_window[ii]

    return kSpace_hanning


class ReconstructionTabController(ReconstructionTabWidget):
    """
    Controller class for the ReconstructionTabWidget.

    Inherits from ReconstructionTabWidget to provide additional functionality for image reconstruction.

    Attributes:
        pocs_button: QPushButton for performing POCS.
        image_fft_button: QPushButton for performing FFT reconstruction.
        image_art_button: QPushButton for performing ART reconstruction.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the ReconstructionTabController.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(ReconstructionTabController, self).__init__(*args, **kwargs)

        # Connect the image_fft_button clicked signal to the fftReconstruction method
        self.pocs_button.clicked.connect(self.pocsReconstruction)
        self.zero_button.clicked.connect(self.zeroReconstruction)
        self.ifft_button.clicked.connect(self.ifft)
        self.dfft_button.clicked.connect(self.dfft)
        self.image_art_button.clicked.connect(self.artReconstruction)

    def dfft(self):

        thread = threading.Thread(target=self.runDFFT)
        thread.start()

    def runDFFT(self):
        """
        """
        # Get the k-space data from the main matrix
        image = self.main.image_view_widget.main_matrix

        # Perform direct FFT shift, inverse FFT, and inverse FFT shift to reconstruct the image in the spatial domain
        k_space = np.fft.fftshift(np.fft.fftn(np.fft.fftshift(image)))

        # Update the main matrix of the image view widget with the image fft data
        self.main.image_view_widget.main_matrix = k_space
        orientation=None
        if self.main.toolbar_image.mat_data and 'axesOrientation' in self.main.toolbar_image.mat_data:
            orientation = self.main.toolbar_image.mat_data['axesOrientation'][0]
        # Add new item to the history list
        self.main.history_list.addNewItem(stamp="dFFT",
                                          image=self.main.image_view_widget.main_matrix,
                                          orientation=orientation,
                                          operation="dFFT",
                                          space="k",
                                          image_key=self.main.image_view_widget.image_key)

    def ifft(self):
        """
        Perform FFT reconstruction in a separate thread.

        Creates a new thread and runs the runFftReconstruction method in that thread.
        """

        thread = threading.Thread(target=self.runIFFT)
        thread.start()

    def runIFFT(self):
        """
        Perform FFT reconstruction.

        Retrieves the k-space data from the main matrix of the image view widget.
        Performs inverse FFT shift, inverse FFT, and inverse FFT shift to reconstruct the image in the spatial domain.
        Updates the main matrix of the image view widget with the reconstructed image.
        Adds the "FFT" operation to the history widget and updates the history dictionary and operations history.
        """
        # Get the k-space data from the main matrix
        k_space = self.main.image_view_widget.main_matrix

        # Perform inverse FFT shift, inverse FFT, and inverse FFT shift to reconstruct the image in the spatial domain
        image = np.fft.ifftshift(np.fft.ifftn(np.fft.ifftshift(k_space)))

        # Update the main matrix of the image view widget with the image fft data
        self.main.image_view_widget.main_matrix = image

        figure = image / np.max(np.abs(image)) * 100
        
        orientation = None
        if self.main.toolbar_image.mat_data and 'axesOrientation' in self.main.toolbar_image.mat_data:
            orientation = self.main.toolbar_image.mat_data['axesOrientation'][0]

        # Add new item to the history list
        self.main.history_list.addNewItem(stamp="iFFT",
                                          image=figure,
                                          orientation=orientation,
                                          operation="iFFT",
                                          space="i",
                                          image_key=self.main.image_view_widget.image_key)

    def artReconstruction(self):
        """
        Perform ART reconstruction in a separate thread.

        Creates a new thread and runs the runArtReconstruction method in that thread.
        """

        thread = threading.Thread(target=self.runArtReconstruction)
        thread.start()

    def runArtReconstruction(self):
        """
        Perform ART reconstruction.

        Retrieves the mat data from the loaded .mat file in the main toolbar controller.
        Extracts the necessary data from the mat file.
        Performs the ART reconstruction algorithm.
        Updates the main matrix of the image view widget with the reconstructed image.
        Adds the "ART" operation to the history widget and updates the history dictionary and operations history.
        """
        # Get the mat data from the loaded .mat file in the main toolbar controller
        mat_data = self.main.toolbar_image.mat_data

        # Extract datas data from the loaded .mat file
        sampled = self.main.toolbar_image.k_space_raw
        fov = np.reshape(mat_data['fov'], -1) * 1e-2
        nPoints = np.reshape(mat_data['nPoints'], -1)
        k = sampled[:, 0:3]
        s = sampled[:, 3]

        # Points where rho will be estimated
        x = np.linspace(-fov[0] / 2, fov[0] / 2, nPoints[0])
        y = np.linspace(-fov[1] / 2, fov[1] / 2, nPoints[1])
        z = np.linspace(-fov[2] / 2, fov[2] / 2, nPoints[2])
        y, z, x = np.meshgrid(y, z, x)
        x = np.reshape(x, (-1, 1))
        y = np.reshape(y, (-1, 1))
        z = np.reshape(z, (-1, 1))

        # k-points
        kx = np.reshape(k[:, 0], (-1, 1))
        ky = np.reshape(k[:, 1], (-1, 1))
        kz = np.reshape(k[:, 2], (-1, 1))

        # Iterative process
        lbda = float(self.lambda_text_field.text())
        n_iter = int(self.niter_text_field.text())
        index = np.arange(len(s))

        def iterative_process_gpu(kx, ky, kz, x, y, z, s, rho, lbda, n_iter, index):
            n = 0
            n_samples = len(s)
            m = 0
            for iteration in range(n_iter):
                cp.random.shuffle(index)
                for jj in range(n_samples):
                    ii = index[jj]
                    x0 = cp.exp(-1j * 2 * cp.pi * (kx[ii] * x + ky[ii] * y + kz[ii] * z))
                    x1 = (x0.T @ rho) - s[ii]
                    x2 = x1 * cp.conj(x0) / (cp.conj(x0.T) @ x0)
                    d_rho = lbda * x2
                    rho -= d_rho
                    n += 1
                    if n / n_samples > 0.01:
                        m += 1
                        n = 0
                        print("ART iteration %i: %i %%" % (iteration + 1, m))

            return rho

        def iterative_process_cpu(kx, ky, kz, x, y, z, s, rho, lbda, n_iter, index):
            n = 0
            n_samples = len(s)
            m = 0
            for iteration in range(n_iter):
                np.random.shuffle(index)
                for jj in range(n_samples):
                    ii = index[jj]
                    x0 = np.exp(-1j * 2 * np.pi * (kx[ii] * x + ky[ii] * y + kz[ii] * z))
                    x1 = (x0.T @ rho) - s[ii]
                    x2 = x1 * np.conj(x0) / (np.conj(x0.T) @ x0)
                    d_rho = lbda * x2
                    rho -= d_rho
                    n += 1
                    if n / n_samples > 0.01:
                        m += 1
                        n = 0
                        print("ART iteration %i: %i %%" % (iteration + 1, m))

            return rho

        # Launch the GPU function
        rho = np.reshape(np.zeros((nPoints[0] * nPoints[1] * nPoints[2]), dtype=complex), (-1, 1))
        start = time.time()

        gpu_ready = False
        if 'cp' in globals():
            try:
                print('Executing ART in GPU...')

                # Transfer numpy arrays to cupy arrays
                kx_gpu = cp.asarray(kx)
                ky_gpu = cp.asarray(ky)
                kz_gpu = cp.asarray(kz)
                x_gpu = cp.asarray(x)
                y_gpu = cp.asarray(y)
                z_gpu = cp.asarray(z)
                s_gpu = cp.asarray(s)
                index = cp.asarray(index)
                rho_gpu = cp.asarray(rho)

                # Execute ART
                rho_gpu = iterative_process_gpu(kx_gpu, ky_gpu, kz_gpu, x_gpu, y_gpu, z_gpu, s_gpu, rho_gpu, lbda, n_iter,
                                                index)
                rho = cp.asnumpy(rho_gpu)
                gpu_ready = True
            except:
                print("GPU not available...")
        if not gpu_ready:
            print('Executing ART in CPU...')

            rho = iterative_process_cpu(kx, ky, kz, x, y, z, s, rho, lbda, n_iter,
                                            index)
        end = time.time()
        print("Reconstruction time = %0.1f s" % (end - start))

        rho = np.reshape(rho, nPoints[-1::-1])

        # Update the main matrix of the image view widget with the image fft data
        self.main.image_view_widget.main_matrix = rho

        figure = rho/np.max(np.abs(rho))*100
        orientation=None
        if self.main.toolbar_image.mat_data and 'axesOrientation' in self.main.toolbar_image.mat_data:
            orientation = self.main.toolbar_image.mat_data['axesOrientation'][0]
        # Add new item to the history list
        self.main.history_list.addNewItem(stamp="ART",
                                          image=figure,
                                          orientation=orientation,
                                          operation="ART n = %i, lambda = %0.3f" % (n_iter, lbda),
                                          space="i",
                                          image_key=self.main.image_view_widget.image_key)

        return

    def zeroReconstruction(self):

        thread = threading.Thread(target=self.runZeroReconstruction)
        thread.start()

    def runZeroReconstruction(self):
        """
        Run the partial reconstruction operation.

        Retrieves the necessary parameters and performs the partial reconstruction on the loaded image.
        Updates the main matrix of the image view widget with the partially reconstructed image, adds the operation to
        the history widget, and updates the operations history.
        """
        # Get the k_space data and its shape
        k_space = self.main.image_view_widget.main_matrix.copy()
        img_ref = np.abs(np.fft.ifftshift(np.fft.ifftn(np.fft.ifftshift(k_space))))
        nPoints = self.main.toolbar_image.nPoints[-1::-1]

        # Percentage for partial reconstruction from the text field
        factors = self.partial_reconstruction_factor.text().split(',')
        factors = [float(num) for num in factors][-1::-1]
        mm = np.array([round(num) for num in (nPoints * factors)])

        # Set to zero the corresponding values
        k_space[mm[0]::, :, :] = 0.0
        k_space[:, mm[1]::, :] = 0.0
        k_space[:, :, mm[2]::] = 0.0

        # Calculate logarithmic scale
        image = np.abs(np.fft.ifftshift(np.fft.ifftn(np.fft.ifftshift(k_space))))

        # Get correlation with reference image
        correlation = np.corrcoef(img_ref.flatten(), image.flatten())[0, 1]
        print("Respect the reference image:")
        print("Convergence: %0.2e" % (1 - correlation))

        # Update the main matrix of the image view widget with the k-space data
        self.main.image_view_widget.main_matrix = image
        orientation=None
        if self.main.toolbar_image.mat_data and 'axesOrientation' in self.main.toolbar_image.mat_data:
            orientation = self.main.toolbar_image.mat_data['axesOrientation'][0]
        # Add new item to the history list
        self.main.history_list.addNewItem(stamp="Partial Zero Reconstruction",
                                          image=image,
                                          orientation=orientation,
                                          operation="Partial Reconstruction - " + str(factors[-1::-1]),
                                          space="i",
                                          image_key=self.main.image_view_widget.image_key)

    def pocsReconstruction(self):
        """
        Perform POCS reconstruction in a separate thread.

        Creates a new thread and runs the runPocsReconstruction method in that thread.
        """

        thread = threading.Thread(target=self.runPocsReconstruction)
        thread.start()

    def runPocsReconstruction(self):
        # Get n_points
        mat_data = self.main.toolbar_image.mat_data
        n_points = mat_data['nPoints'][0][-1::-1]

        # Number of extra lines which has been taken past the center of k-space
        factors = self.partial_reconstruction_factor.text().split(',')
        factors = [float(num) for num in factors][-1::-1]

        # Get the k_space data
        k_space_ref = self.main.image_view_widget.main_matrix.copy()

        # Run pocs
        img_reconstructed = utils.run_pocs_reconstruction(n_points, factors, k_space_ref)

        # Update the main matrix of the image view widget with the interpolated image
        self.main.image_view_widget.main_matrix = img_reconstructed

        figure = img_reconstructed / np.max(np.abs(img_reconstructed)) * 100

        # Add new item to the history list
        orientation=None
        if self.main.toolbar_image.mat_data and 'axesOrientation' in self.main.toolbar_image.mat_data:
            orientation = self.main.toolbar_image.mat_data['axesOrientation'][0]
        self.main.history_list.addNewItem(stamp="POCS",
                                          image=figure,
                                          orientation=orientation,
                                          operation="POCS - " + str(factors[-1::-1]),
                                          space="i",
                                          image_key=self.main.image_view_widget.image_key)



#!/usr/bin/env python3
# Basic hacks and tests for marga system
#

import numpy as np
import matplotlib.pyplot as plt

import socket, time
from local_config import ip_address, port, grad_board, fpga_clk_freq_MHz
from server_comms import *

from marmachine import *
import marcompile as fc

import pdb
st = pdb.set_trace

def get_exec_state(socket, display=True):
    reply, status = command({'regstatus': 0}, socket, print_infos=True)
    exec_reg = reply[4]['regstatus'][0]
    state = exec_reg >> 24
    pc = exec_reg & 0xffffff
    if display:
        print('state: {:d}, PC: {:d}'.format(state, pc))
    return state, pc

def run_manual_test(data, interval=0.001, timeout=20):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.connect((ip_address, port))
        packet_idx = 0

        # mem data
        # data = np.zeros(65536, dtype=np.uint32)
        command( {'mar_mem': data.tobytes()} , s)

        # Start execution
        command({'ctrl': 0x1}, s, print_infos=True)

        for k in range(int(timeout/interval)):
            time.sleep(interval)
            state, pc = get_exec_state(s)
            if (state == STATE_HALT):
                # one final acquisition to kill time on the simulation
                get_exec_state(s)
                break

        # Stop execution
        command({'ctrl': 0x2}, s, print_infos=True)

        # Flush RX data
        rx_data = command ({'flush_rx':0}, s, print_infos=True)[0][4]['flush_rx']

        # Auto-close server if it's simulating
        if command({'are_you_real':0}, s)[0][4]['are_you_real'] == "simulation":
            send_packet(construct_packet({}, 0, command=close_server_pkt), s)

        return rx_data

def run_streaming_test(data):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.connect((ip_address, port))
        packet_idx = 0

        if False: # clear memory first
            command( {'mar_mem': np.zeros(65536, dtype=np.uint32).tobytes()} , s)

        # mem data
        rx_data, msgs = command({'run_seq': data.tobytes()} , s)

        # print(msgs)

        # Auto-close server if it's simulating
        if command({'are_you_real':0}, s)[0][4]['are_you_real'] == "simulation":
            send_packet(construct_packet({}, 0, command=close_server_pkt), s)

        return rx_data

def leds():
    time_interval_s = 4 # how long the total sequence should run for
    total_states = 256
    delay = int(np.round(time_interval_s * fpga_clk_freq_MHz * 1e6 / total_states))

    raw_data = np.zeros(65536, dtype=np.uint32)
    addr = 0
    for k in range(256):
        raw_data[addr] = insta(IWAIT, delay - 1); addr += 1 # -1 because the next instruction takes 1 cycle
        # raw_data[addr] = insta(IWAIT, 0x300000); addr += 1 # -1 because the next instruction takes 1 cycle
        raw_data[addr] = instb(GATES_LEDS, 0, (k & 0xff) << 8); addr += 1
    # go idle
    raw_data[addr] = insta(IFINISH, 0); addr += 1
    raw_data = raw_data[:addr] # truncate
    return raw_data

def marcompile_test():

    lc = fc.csv2bin("/tmp/mar_test1.csv")
    lc.append(insta(IFINISH, 0))
    raw_data = np.array(lc, dtype=np.uint32)
    print(raw_data.size)
    return raw_data

# Utility function to generate RX control words instead of doing it inline
def rx_ctrl(delay, rx0_en, rx1_en, rx0_rate_valid, rx1_rate_valid,
            rx0_lo, rx1_lo, rx0_rst_n=True, rx1_rst_n=True):
    return instb( RX_CTRL, delay,
                  (rx1_en << 9) | (rx0_en << 8) | (rx1_rst_n << 7) | (rx0_rst_n << 6) \
                  | (rx1_rate_valid << 5) | (rx0_rate_valid << 4) \
                  | (rx1_lo << 2) | (rx0_lo << 0) )

def example_tr_loop():
    lo_freq0 = 5 # MHz
    lo_freq1 = 6 # MHz
    lo_amp = 100 # percent
    raw_data = np.zeros(1000000, dtype=np.uint32)
    addr = 0

    cic0_decimation = 100
    cic1_decimation = 200
    dds_demod_ch = 3

    tr_loops = 2

    # Initially, turn on LO
    dds0_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq0).astype(np.uint32) # 31b phase accumulator in marga
    dds1_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq1).astype(np.uint32) # 31b phase accumulator in marga
    assert dds0_phase_step < 2**31, "DDS frequency outside of valid range"
    assert dds1_phase_step < 2**31, "DDS frequency outside of valid range"

    # zero the phase increment initially and reset the phase
    raw_data[addr] = instb(DDS0_PHASE_LSB, 3, 0); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 2, 0x8000); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_LSB, 1, 0); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_MSB, 0, 0x8000); addr += 1

    # set the phase increment, start phase going
    raw_data[addr] = instb(DDS0_PHASE_LSB, 3, dds0_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 2, dds0_phase_step >> 16); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_LSB, 1, dds1_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_MSB, 0, dds1_phase_step >> 16); addr += 1

    # allow the new phase to propagate through the chain before setting nonzero I/Q values
    raw_data[addr] = insta(IWAIT, 100); addr += 1 # max I value

    # configure RX settings: both channels to use DDS source 0
    # reset CICs
    raw_data[addr] = rx_ctrl(0, 0, 0, 0, 0, 0, 0, 0, 0); addr += 1
    # take them out of reset
    raw_data[addr] = rx_ctrl(40, 0, 0, 0, 0, dds_demod_ch, dds_demod_ch, 1, 1); addr += 1
    # set control buses to new rates (immediately for now)
    raw_data[addr] = instb(RX0_RATE, 0, cic0_decimation); addr += 1
    raw_data[addr] = instb(RX1_RATE, 0, cic1_decimation); addr += 1

    # briefly signal that there's a new rate
    raw_data[addr] = rx_ctrl(40, 0, 0, 1, 1, dds_demod_ch, dds_demod_ch); addr += 1
    raw_data[addr] = rx_ctrl(0, 0, 0, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1

    for k in range(tr_loops):

        ## Initial TX, hard pulse
        wait = 1.6276 # 1.6276us will lead to exactly 200 cycles

        initial_block_delay = 4

        # initial delay of 100 cycles, taking into account the next block
        raw_data[addr] = insta(IWAIT, 100 - initial_block_delay - 5); addr += 1

        addr_tx_start = addr
        raw_data[addr] = instb(GATES_LEDS, initial_block_delay + 1, 0x1); addr += 1 # turn on TX gate
        raw_data[addr] = instb(TX0_I, initial_block_delay - 1, 0xc000); addr += 1
        raw_data[addr] = instb(TX0_Q, initial_block_delay - 2, 0x3fff); addr += 1
        raw_data[addr] = instb(TX1_I, initial_block_delay - 3, 0xc000); addr += 1
        raw_data[addr] = instb(TX1_Q, initial_block_delay - 4, 0x3fff); addr += 1

        # take into account the time offset due to internal instructions and pre/post-wait instruction delays
        # Written assuming the TX and TX gate outputs should be perfectly in sync - in reality a delay is probably wise
        initial_block_delay = 4
        raw_data[addr] = insta(IWAIT, int(np.round(wait * fpga_clk_freq_MHz)) + 1 - initial_block_delay - addr + addr_tx_start); addr += 1
        raw_data[addr] = instb(GATES_LEDS, initial_block_delay + 1, 0x0); addr += 1; # turn off TX gate
        raw_data[addr] = instb(TX0_I, initial_block_delay - 1, 0x0); addr += 1
        raw_data[addr] = instb(TX0_Q, initial_block_delay - 2, 0x0); addr += 1
        raw_data[addr] = instb(TX1_I, initial_block_delay - 3, 0x0); addr += 1
        raw_data[addr] = instb(TX1_Q, initial_block_delay - 4, 0x0); addr += 1
        addr_tx_end = addr

        # Wait then acquire some data, wait some more, stop acquisition
        raw_data[addr] = insta(IWAIT, 100 - 4 - addr + addr_tx_end); addr += 1 # wait 100 cycles
        raw_data[addr] = rx_ctrl(40, 1, 1, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1
        raw_data[addr] = insta(IWAIT, 500 - 5); addr += 1 # wait 500 cycles
        raw_data[addr] = rx_ctrl(40, 0, 0, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1

    # Go idle
    raw_data[addr] = insta(IFINISH, 0); addr += 1
    raw_data = raw_data[:addr] # truncate
    print(addr)
    return raw_data

def tx_short():
    lo_freq0 = 5 # MHz
    lo_freq1 = 10
    lo_freq2 = 1.5
    lo_freq3 = 13.333333
    lo_amp = 100 # percent

    raw_data = np.zeros(65536, dtype=np.uint32)
    addr = 0

    # Turn on LO
    dds0_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq0).astype(np.uint32) # 31b phase accumulator in marga
    dds1_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq1).astype(np.uint32) # 31b phase accumulator in marga
    assert dds0_phase_step < 2**31, "DDS frequency outside of valid range"
    assert dds1_phase_step < 2**31, "DDS frequency outside of valid range"

    # zero the phase increment initially and reset the phase
    raw_data[addr] = instb(DDS0_PHASE_LSB, 3, 0); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 2, 0x8000); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_LSB, 1, 0); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_MSB, 0, 0x8000); addr += 1

    # set the phase increment, start phase going
    raw_data[addr] = instb(DDS0_PHASE_LSB, 3, dds0_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 2, dds0_phase_step >> 16); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_LSB, 1, dds1_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_MSB, 0, dds1_phase_step >> 16); addr += 1

    # allow the new phase to propagate through the chain before setting nonzero I/Q values
    raw_data[addr] = insta(IWAIT, 20); addr += 1 # max I value

    steps = 300
    pattern0 = np.hstack([ np.linspace(1, 0, steps//3), np.linspace(1, -1, steps//3), np.linspace(-1, 1, steps//3) ])
    pattern1 = np.hstack([ np.linspace(0, 1, steps//3), np.linspace(1, 0, steps//3), np.linspace(0, -1, steps//3) ])
    tdata0 = np.round(0x7fff * pattern0).astype(np.int16)
    tdata1 = np.round(0x7fff * pattern1).astype(np.int16)

    for k in range(steps):
        hv0 = tdata0[k] & 0xffff
        hv1 = tdata1[k] & 0xffff
        raw_data[addr] = instb(TX0_I, 3, hv0); addr += 1
        raw_data[addr] = instb(TX0_Q, 2, hv0); addr += 1
        raw_data[addr] = instb(TX1_I, 1, hv1); addr += 1
        raw_data[addr] = instb(TX1_Q, 0, hv1); addr += 1

    # do mid-scale output, and change frequency
    raw_data[addr] = instb(TX0_I, 3, 0x4000); addr += 1
    raw_data[addr] = instb(TX0_Q, 2, 0x4000); addr += 1
    raw_data[addr] = instb(TX1_I, 1, 0x4000); addr += 1
    raw_data[addr] = instb(TX1_Q, 0, 0x4000); addr += 1

    # wait for 3us; shortened delay by 4 cycles for the next instructions to be right on time
    raw_data[addr] = insta(IWAIT, int(3 * fpga_clk_freq_MHz) - 4); addr += 1

    dds2_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq2).astype(np.uint32) # 31b phase accumulator in marga
    dds3_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq3).astype(np.uint32) # 31b phase accumulator in marga

    # switch frequency on both channels simultaneously, no reset
    raw_data[addr] = instb(DDS0_PHASE_LSB, 3, dds2_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 2, dds2_phase_step >> 16); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_LSB, 1, dds3_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_MSB, 0, dds3_phase_step >> 16); addr += 1

    raw_data[addr] = insta(IWAIT, int(4 * fpga_clk_freq_MHz)); addr += 1

    # go idle
    raw_data[addr] = instb(TX0_I, 3, 0); addr += 1
    raw_data[addr] = instb(TX0_Q, 2, 0x8000); addr += 1
    raw_data[addr] = instb(TX1_I, 1, 0); addr += 1
    raw_data[addr] = instb(TX1_Q, 0, 0x8000); addr += 1
    raw_data[addr] = insta(IFINISH, 0); addr += 1
    raw_data = raw_data[:addr] # truncate
    return raw_data

def rx_short():
    raw_data = np.zeros(65536, dtype=np.uint32)
    addr = 0

    # turn DDS source 0 on
    lo_freq0 = 5 # MHz
    cic_decimation = 4
    dds_demod_ch = 0
    acquisition_ticks = 800
    dds0_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq0).astype(np.uint32) # 31b phase accumulator in marga
    assert dds0_phase_step < 2**31, "DDS frequency outside of valid range"
    raw_data[addr] = instb(DDS0_PHASE_LSB, 1, dds0_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 0, dds0_phase_step >> 16); addr += 1

    # configure RX settings: both channels to use DDS source 3 (DC), decimation of 10
    # reset CICs
    ## raw_data[addr] = instb(RX0_CTRL, 0, 0x0000); addr += 1
    raw_data[addr] = rx_ctrl(0, 0, 0, 0, 0, 0, 0); addr += 1
    # raw_data[addr] = instb(RX1_CTRL, 0, 0x0000); addr += 1
    # take them out of reset later
    ## raw_data[addr] = instb(RX0_CTRL, 50, 0x8000 | cic_decimation | (dds_demod_ch << 12) ); addr += 1
    raw_data[addr] = rx_ctrl(50, 1, 0, 0, 0, dds_demod_ch, 0); addr += 1
    raw_data[addr] = instb(RX0_RATE, 0, cic_decimation); addr += 1
    # briefly signal that there's a new rate
    ## raw_data[addr] = instb(RX0_CTRL, 50, 0xc000 | cic_decimation | (dds_demod_ch << 12) ); addr += 1
    raw_data[addr] = rx_ctrl(50, 1, 0, 1, 0, dds_demod_ch, 0); addr += 1
    # raw_data[addr] = instb(RX1_CTRL, 49, 0xf000 | cic_decimation); addr += 1
    # end the new rate flag (the buffers are not empty so no offset time is needed)
    ## raw_data[addr] = instb(RX0_CTRL, 0, 0x8000 | cic_decimation | (dds_demod_ch << 12) ); addr += 1 # decimation may not be needed here
    raw_data[addr] = rx_ctrl(0, 1, 0, 0, 0, dds_demod_ch, 0); addr += 1
    # raw_data[addr] = instb(RX1_CTRL, 0, 0xb000 | cic_decimation); addr += 1

    # just acquire data for a while
    raw_data[addr] = insta(IWAIT, acquisition_ticks - 1); addr += 1

    # reset RX again
    ## raw_data[addr] = instb(RX0_CTRL, 1, 0x3000); addr += 1
    raw_data[addr] = rx_ctrl(1, 0, 0, 0, 0, dds_demod_ch, 0); addr += 1
    # raw_data[addr] = instb(RX1_CTRL, 0, 0x3000); addr += 1

    # turn off DDS0
    raw_data[addr] = instb(DDS0_PHASE_LSB, 1, 0); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 0, 0); addr += 1

    # go idle
    raw_data[addr] = insta(IFINISH, 0); addr += 1
    raw_data = raw_data[:addr] # truncate
    return raw_data

def loopback(cic0_decimation=7, cic1_decimation=10):
    lo_freq0 = 4 # MHz
    lo_freq1 = 4
    lo_freq2 = 1.5
    lo_freq3 = 1.5
    lo_amp = 100 # percent
    dds_demod_ch = 0

    extra_time = 20

    raw_data = np.zeros(65536, dtype=np.uint32)
    addr = 0

    # Turn on LO
    dds0_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq0).astype(np.uint32) # 31b phase accumulator in marga
    dds1_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq1).astype(np.uint32) # 31b phase accumulator in marga
    assert dds0_phase_step < 2**31, "DDS frequency outside of valid range"
    assert dds1_phase_step < 2**31, "DDS frequency outside of valid range"

    # zero the phase increment initially and reset the phase
    raw_data[addr] = instb(DDS0_PHASE_LSB, 7, 0); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 6, 0x8000); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_LSB, 5, 0); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_MSB, 4, 0x8000); addr += 1

    # set the phase increment, start phase going
    raw_data[addr] = instb(DDS0_PHASE_LSB, 0, dds0_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 0, dds0_phase_step >> 16); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_LSB, 0, dds1_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_MSB, 0, dds1_phase_step >> 16); addr += 1

    # allow the new phase to propagate through the chain before setting nonzero I/Q values
    raw_data[addr] = insta(IWAIT, 100); addr += 1 # max I value

    # configure RX settings: both channels to use DDS source 3 (DC), decimation of 10
    # reset CICs
    raw_data[addr] = rx_ctrl(0, 0, 0, 0, 0, dds_demod_ch, dds_demod_ch, 0, 0); addr += 1
    # raw_data[addr] = instb(RX1_CTRL, 0, 0x0000); addr += 1
    # take them out of reset later
    raw_data[addr] = rx_ctrl(40, 0, 0, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1
    raw_data[addr] = instb(RX0_RATE, 0, cic0_decimation); addr += 1
    raw_data[addr] = instb(RX1_RATE, 0, cic1_decimation); addr += 1
    # briefly signal that there's a new rate
    raw_data[addr] = rx_ctrl(40, 0, 0, 1, 1, dds_demod_ch, dds_demod_ch); addr += 1
    # raw_data[addr] = instb(RX1_CTRL, 49, 0xf000 | cic_decimation); addr += 1
    # end the new rate flag and start acquisition (the buffers are not empty so no offset time is needed)
    raw_data[addr] = rx_ctrl(0, 1, 1, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1

    # Wait a bit, so that all buffers are emptied
    raw_data[addr] = insta(IWAIT, 100); addr += 1 # max I value

    if False:
        ## NOTE: example of how to send arbitrary data to RX0 or RX1 CIC:
        # first, set the output bus to the data you want to send (in this case, cic0_decimation)
        raw_data[addr] = instb(RX0_RATE, 0, cic0_decimation); addr += 1
        # next, strobe the 'bus 0 valid' line of the RX control buffer (no delays needed if it's not the Xilinx CIC)
        raw_data[addr] = rx_ctrl(0, 0, 0, 1, 0, dds_demod_ch, dds_demod_ch); addr += 1
        raw_data[addr] = rx_ctrl(0, 0, 0, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1
        # same for channel 1:
        raw_data[addr] = instb(RX1_RATE, 0, cic1_decimation); addr += 1
        raw_data[addr] = rx_ctrl(0, 0, 0, 0, 1, dds_demod_ch, dds_demod_ch); addr += 1
        raw_data[addr] = rx_ctrl(0, 0, 0, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1
        # wait 100 cycles and reset both CICs, then wait 100 more and take them out of reset and start acquisition
        raw_data[addr] = rx_ctrl(100, 0, 0, 0, 0, dds_demod_ch, dds_demod_ch, 0, 0); addr += 1
        raw_data[addr] = rx_ctrl(100, 1, 1, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1

        # Wait a bit, so that all buffers are emptied
        raw_data[addr] = insta(IWAIT, 100); addr += 1 # max I value

    steps = 300
    pattern0 = np.hstack([ np.linspace(1, 0, steps//3), np.linspace(1, -1, steps//3), np.linspace(-1, 1, steps//3) ])
    # pattern1 = np.hstack([ np.linspace(0, 1, steps//3), np.linspace(1, 0, steps//3), np.linspace(0, -1, steps//3) ])
    pattern1 = pattern0
    tdata0 = np.round(0x7fff * pattern0).astype(np.int16)
    tdata1 = np.round(0x7fff * pattern1).astype(np.int16)

    for k in range(steps):
        hv0 = tdata0[k] & 0xffff
        hv1 = tdata1[k] & 0xffff
        raw_data[addr] = instb(TX0_I, 3, hv0); addr += 1
        raw_data[addr] = instb(TX0_Q, 2, hv0); addr += 1
        raw_data[addr] = instb(TX1_I, 1, hv1); addr += 1
        raw_data[addr] = instb(TX1_Q, 0, hv1); addr += 1

    # do mid-scale output, and change frequency
    raw_data[addr] = instb(TX0_I, 3, 0x4000); addr += 1
    raw_data[addr] = instb(TX0_Q, 2, 0x4000); addr += 1
    raw_data[addr] = instb(TX1_I, 1, 0x4000); addr += 1
    raw_data[addr] = instb(TX1_Q, 0, 0x4000); addr += 1

    # wait for 3us; shortened delay by 4 cycles for the next instructions to be right on time
    raw_data[addr] = insta(IWAIT, int(3 * fpga_clk_freq_MHz) - 4); addr += 1

    dds2_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq2).astype(np.uint32) # 31b phase accumulator in marga
    dds3_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq3).astype(np.uint32) # 31b phase accumulator in marga

    # switch frequency on both channels simultaneously, no reset
    raw_data[addr] = instb(DDS0_PHASE_LSB, 3, dds2_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 2, dds2_phase_step >> 16); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_LSB, 1, dds3_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_MSB, 0, dds3_phase_step >> 16); addr += 1

    # wait for extra acquisition time
    raw_data[addr] = insta(IWAIT, int(extra_time * fpga_clk_freq_MHz)); addr += 1

    # end RX
    raw_data[addr] = rx_ctrl(0, 0, 0, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1

    # go idle
    raw_data[addr] = instb(TX0_I, 3, 0); addr += 1
    raw_data[addr] = instb(TX0_Q, 2, 0x0000); addr += 1
    raw_data[addr] = instb(TX1_I, 1, 0); addr += 1
    raw_data[addr] = instb(TX1_Q, 0, 0x0000); addr += 1
    raw_data[addr] = insta(IFINISH, 0); addr += 1
    # raw_data = raw_data[:addr] # truncate
    return raw_data

def long_loopback():
    lo_freq0 = 4 # MHz
    lo_freq1 = 4.01
    lo_freq2 = 1.5
    lo_freq3 = 1.51
    lo_amp = 100 # percent

    # These settings lead to a fairly normal sequence
    if False:
        cic0_decimation = 1000
        cic1_decimation = 500
        extra_time = 20
        sine_ts = 2000
        max_addr = 400000

    # These settings lead to memory-buffer-low events
    if False:
        cic0_decimation = 50
        cic1_decimation = 94
        extra_time = 20
        sine_ts = 5
        max_addr = 141000

    # These settings lead to a very long sequence, with a few memory buffer low warnings
    if False:
        cic0_decimation = 100
        cic1_decimation = 200
        extra_time = 20
        sine_ts = 10
        max_addr = 500000

    dds_demod_ch = 0
    raw_data = np.zeros(max_addr, dtype=np.uint32) # massive sequence
    addr = 0

    # Turn on LO
    dds0_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq0).astype(np.uint32) # 31b phase accumulator in marga
    dds1_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq1).astype(np.uint32) # 31b phase accumulator in marga
    assert dds0_phase_step < 2**31, "DDS frequency outside of valid range"
    assert dds1_phase_step < 2**31, "DDS frequency outside of valid range"

    # zero the phase increment initially and reset the phase
    raw_data[addr] = instb(DDS0_PHASE_LSB, 7, 0); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 6, 0x8000); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_LSB, 5, 0); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_MSB, 4, 0x8000); addr += 1

    # set the phase increment, start phase going
    raw_data[addr] = instb(DDS0_PHASE_LSB, 0, dds0_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 0, dds0_phase_step >> 16); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_LSB, 0, dds1_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_MSB, 0, dds1_phase_step >> 16); addr += 1

    # allow the new phase to propagate through the chain before setting nonzero I/Q values
    raw_data[addr] = insta(IWAIT, 100); addr += 1 # max I value

    # configure RX settings: both channels to use DDS source 3 (DC), decimation of 10
    # reset CICs
    raw_data[addr] = rx_ctrl(0, 0, 0, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1
    # raw_data[addr] = instb(RX1_CTRL, 0, 0x0000); addr += 1
    # take them out of reset later
    raw_data[addr] = rx_ctrl(40, 1, 1, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1
    raw_data[addr] = instb(RX0_RATE, 0, cic0_decimation); addr += 1
    raw_data[addr] = instb(RX1_RATE, 0, cic1_decimation); addr += 1
    # briefly signal that there's a new rate
    raw_data[addr] = rx_ctrl(40, 1, 1, 1, 1, dds_demod_ch, dds_demod_ch); addr += 1
    # raw_data[addr] = instb(RX1_CTRL, 49, 0xf000 | cic_decimation); addr += 1
    # end the new rate flag (the buffers are not empty so no offset time is needed)
    raw_data[addr] = rx_ctrl(0, 1, 1, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1
    # raw_data[addr] = instb(RX1_CTRL, 0, 0xb000 | cic_decimation); addr += 1

    steps = 300
    pattern0 = np.hstack([ np.linspace(1, 0, steps//3), np.linspace(1, -1, steps//3), np.linspace(-1, 1, steps//3) ])
    # pattern1 = np.hstack([ np.linspace(0, 1, steps//3), np.linspace(1, 0, steps//3), np.linspace(0, -1, steps//3) ])
    pattern1 = pattern0
    tdata0 = np.round(0x7fff * pattern0).astype(np.int16)
    tdata1 = np.round(0x7fff * pattern1).astype(np.int16)

    for k in range(steps):
        hv0 = tdata0[k] & 0xffff
        hv1 = tdata1[k] & 0xffff
        raw_data[addr] = instb(TX0_I, 3, hv0); addr += 1
        raw_data[addr] = instb(TX0_Q, 2, hv0); addr += 1
        raw_data[addr] = instb(TX1_I, 1, hv1); addr += 1
        raw_data[addr] = instb(TX1_Q, 0, hv1); addr += 1

    # do mid-scale output, and change frequency
    raw_data[addr] = instb(TX0_I, 3, 0x4000); addr += 1
    raw_data[addr] = instb(TX0_Q, 2, 0x4000); addr += 1
    raw_data[addr] = instb(TX1_I, 1, 0x4000); addr += 1
    raw_data[addr] = instb(TX1_Q, 0, 0x4000); addr += 1

    # wait for 3us; shortened delay by 4 cycles for the next instructions to be right on time
    raw_data[addr] = insta(IWAIT, int(3 * fpga_clk_freq_MHz) - 4); addr += 1

    dds2_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq2).astype(np.uint32) # 31b phase accumulator in marga
    dds3_phase_step = np.round(2**31 / fpga_clk_freq_MHz * lo_freq3).astype(np.uint32) # 31b phase accumulator in marga

    # switch frequency on both channels simultaneously, no reset
    raw_data[addr] = instb(DDS0_PHASE_LSB, 3, dds2_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS0_PHASE_MSB, 2, dds2_phase_step >> 16); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_LSB, 1, dds3_phase_step & 0xffff); addr += 1
    raw_data[addr] = instb(DDS1_PHASE_MSB, 0, dds3_phase_step >> 16); addr += 1

    # wait for extra acquisition time
    raw_data[addr] = insta(IWAIT, int(extra_time * fpga_clk_freq_MHz)); addr += 1

    ### Long sinusoid of slowly increasing frequency
    start_addr = addr
    end_addr = max_addr - 1000 # reserve 1000 spots at the end for final commands
    phase_points = (end_addr - start_addr) // 5
    t = np.linspace(0, 1, phase_points)
    c = np.cos(2*np.pi*20*t) * (0.2 + 0.5*t)
    s = np.sin(2*np.pi*13*t) * (0.1 + 0.8*t)
    cb = np.round(c * 0x7fff).astype(np.int16) & 0xffff
    sb = np.round(s * 0x7fff).astype(np.int16) & 0xffff

    for k in range(phase_points):
        raw_data[addr] = instb(TX0_I, 0, cb[k]); addr += 1
        raw_data[addr] = instb(TX0_Q, 0, sb[k]); addr += 1
        raw_data[addr] = instb(TX1_I, 0, cb[k]); addr += 1
        raw_data[addr] = instb(TX1_Q, 0, sb[k]); addr += 1
        raw_data[addr] = insta(IWAIT, sine_ts); addr += 1

    # end RX
    raw_data[addr] = rx_ctrl(0, 0, 0, 0, 0, dds_demod_ch, dds_demod_ch); addr += 1

    # go idle
    raw_data[addr] = instb(TX0_I, 0, 0); addr += 1
    raw_data[addr] = instb(TX0_Q, 0, 0x8000); addr += 1
    raw_data[addr] = instb(TX1_I, 0, 0); addr += 1
    raw_data[addr] = instb(TX1_Q, 0, 0x8000); addr += 1
    raw_data[addr] = insta(IFINISH, 0); addr += 1
    raw_data = raw_data[:addr] # truncate
    print(addr)
    return raw_data

if __name__ == "__main__":
    if False: # flash the LEDs
        run_streaming_test(leds())

    if False:
        data = long_loopback()
        st()

    if False: # single shot, plot RX
        res = run_streaming_test(example_tr_loop())
        # res = run_streaming_test(marcompile_test())

        rxd = res[4]['run_seq']
        # offsets = 1e8
        offsets = 0
        rx0_i = np.array(rxd['rx0_i']).astype(np.int32)
        rx0_q = np.array(rxd['rx0_q']).astype(np.int32)+offsets
        rx1_i = np.array(rxd['rx1_i']).astype(np.int32)+2*offsets
        rx1_q = np.array(rxd['rx1_q']).astype(np.int32)+3*offsets
        plt.plot(rx0_i)
        plt.plot(rx0_q)
        plt.plot(rx1_i)
        plt.plot(rx1_q)

        plt.legend(['rx0_i', 'rx0_q', 'rx1_i', 'rx1_q'])

        plt.show()

    if True: # single shot loopback

        # res = run_streaming_test(long_loopback())
        res = run_streaming_test(loopback())

        rxd = res[4]['run_seq']
        # offsets = 1e8
        offsets = 0
        rx0_i = np.array(rxd['rx0_i']).astype(np.int32)
        rx0_q = np.array(rxd['rx0_q']).astype(np.int32)
        rx1_i = np.array(rxd['rx1_i']).astype(np.int32)
        rx1_q = np.array(rxd['rx1_q']).astype(np.int32)
        plt.plot(rx0_i)
        plt.plot(rx0_q)
        plt.plot(rx1_i)
        plt.plot(rx1_q)

        plt.legend(['rx0_i', 'rx0_q', 'rx1_i', 'rx1_q'])

        plt.show()

    if False: # multiple trials, verify mean against saved
        verify_means = True
        save_means = False # If true, will overwrite reference data - use with care!

        trials = 100

        cic0_decimation = 6
        cic1_decimation = 11

        res = run_streaming_test(loopback(cic0_decimation, cic1_decimation))
        rxd = res[4]['run_seq']
        rx0_pts = len(rxd['rx0_i'])
        rx1_pts = len(rxd['rx1_i'])

        rx0_i = np.zeros([trials, rx0_pts]).astype(np.int32)
        rx0_q = np.zeros([trials, rx0_pts]).astype(np.int32)
        rx1_i = np.zeros([trials, rx1_pts]).astype(np.int32)
        rx1_q = np.zeros([trials, rx1_pts]).astype(np.int32)

        for k in range(trials):
            res = run_streaming_test(loopback(cic0_decimation, cic1_decimation))
            rxd = res[4]['run_seq']
            # offsets = 1e8
            offsets = 0
            rx0_i[k,:] = np.array(rxd['rx0_i']).astype(np.int32)
            rx0_q[k,:] = np.array(rxd['rx0_q']).astype(np.int32)
            rx1_i[k,:] = np.array(rxd['rx1_i']).astype(np.int32)
            rx1_q[k,:] = np.array(rxd['rx1_q']).astype(np.int32)

        # plt.legend(['rx0_i', 'rx0_q', 'rx1_i', 'rx1_q'])
        rx0_x = np.arange(rx0_pts) * cic0_decimation / fpga_clk_freq_MHz
        rx1_x = np.arange(rx1_pts) * cic1_decimation / fpga_clk_freq_MHz

        props = {'alpha': 0.3}

        plt.fill_between(rx0_x, rx0_i.min(0), rx0_i.max(0), **props)
        plt.fill_between(rx0_x, rx0_q.min(0), rx0_q.max(0), **props)
        plt.fill_between(rx1_x, rx1_i.min(0), rx1_i.max(0), **props)
        plt.fill_between(rx1_x, rx1_q.min(0), rx1_q.max(0), **props)

        rx0im, rx0qm, rx1im, rx1qm = rx0_i.mean(0), rx0_q.mean(0), rx1_i.mean(0), rx1_q.mean(0)
        rx0rms, rx1rms = np.sqrt(rx0im**2 + rx0qm**2), np.sqrt(rx1im**2 + rx1qm**2)

        plt.gca().set_prop_cycle(None)

        plt.plot(rx0_x, rx0im)
        plt.plot(rx0_x, rx0qm)
        plt.plot(rx1_x, rx1im)
        plt.plot(rx1_x, rx1qm)

        plt.plot(rx0_x, rx0rms)
        plt.plot(rx1_x, rx1rms)

        plt.xlabel('t (us)')
        plt.legend(['rx0_i', 'rx0_q', 'rx1_i', 'rx1_q', 'rx0_rms', 'rx1_rms'])

        if False:
            plt.plot(rx0_i.std(0))
            plt.plot(rx0_q.std(0))
            plt.plot(rx1_i.std(0))
            plt.plot(rx1_q.std(0))

        ## Verify data
        if verify_means:
            ref_data = np.load('ref_loopback.npz')
            rx0rmsr, rx1rmsr = ref_data['rx0rms'], ref_data['rx1rms']
            # rx0im[10] *= 5 # deliberately stimulate error
            for rx, rxr in zip([rx0rms, rx1rms], [rx0rmsr, rx1rmsr]):
                diff = ((rx/rx.max() - rxr/rxr.max())**2).sum() / rx.size
                if diff > 1e-5:
                    warnings.warn("Normalised loopback data is not as expected!")

        if save_means:
            np.savez_compressed('ref_loopback.npz', rx0rms=rx0rms, rx1rms=rx1rms)

        plt.show()

    if False:
        trials = 1
        for k in range(trials):
            rxd = run_manual_test(loopback(), interval=0.01, timeout=20)
            rx0 = np.array(rxd['ch0']).astype(np.int32)
            rx1 = np.array(rxd['ch1']).astype(np.int32)
            plt.plot(rx0[::2])
            plt.plot(rx0[1::2])
            plt.plot(rx1[::2])
            plt.plot(rx1[1::2])

        plt.xlabel('sample')
        plt.title('loopback, {:d} trials'.format(trials))
        plt.show()



"""
@author:    José Miguel Algarín
@email:     josalggui@i3m.upv.es
@affiliation:MRILab, i3M, CSIC, Valencia, Spain
"""
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QToolBar, QAction


class ProtocolsToolBar(QToolBar):
    def __init__(self, main, *args, **kwargs):
        super(ProtocolsToolBar, self).__init__(*args, **kwargs)
        self.main = main

        # New protocol
        self.action_new_protocol = QAction(QIcon("resources/icons/addProtocol.png"), "New Protocol", self)
        self.action_new_protocol.setStatusTip("Create a new protocol")
        self.addAction(self.action_new_protocol)

        # Del protocol
        self.action_del_protocol = QAction(QIcon("resources/icons/deleteProtocol.png"), "Remove Protocol", self)
        self.action_del_protocol.setStatusTip("Remove a protocol")
        self.addAction(self.action_del_protocol)

        # New sequence
        self.action_new_sequence = QAction(QIcon("resources/icons/addSequence.png"), "New Sequence", self)
        self.action_new_sequence.setStatusTip("Create a new sequence")
        self.addAction(self.action_new_sequence)

        # Del sequence
        self.action_del_sequence = QAction(QIcon("resources/icons/deleteSequence.png"), "Remove Sequence", self)
        self.action_del_sequence.setStatusTip("Remove a sequence from protocol")
        self.addAction(self.action_del_sequence)



"""
:author:    J.M. Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain

"""
import numpy as np

from marge.widgets.widget_plot1d import Plot1DWidget


class Plot1DController(Plot1DWidget):
    """
    1D plot controller class.

    This class extends the `Plot1DWidget` class and serves as a controller for a 1D plot. It initializes the plot with
    the provided data and handles mouse movement events to display information about the data at the cursor position.

    Methods:
        __init__(self, x_data, y_data, legend, x_label, y_label, title): Initialize the Plot1DController instance.
        mouseMoved(self, evt): Handle the mouseMoved event to display information about the data at the cursor position.

    Attributes:
        y_data: The y data for the plot.
        x_data: The x data for the plot.
        x_label: The label for the x axis.
        y_label: The label for the y axis.
        title: The title of the plot.
        lines: A list of LineItems representing the plotted lines.
        plot_item: The PlotItem representing the plot.
        pen: A list of QPen objects representing the line colors.

    """

    def __init__(self,
                 x_data,  # numpy array
                 y_data,  # list of numpy array
                 legend,  # list of strings
                 x_label,  # string
                 y_label,  # string
                 title,  # string
                 ):
        """
        Initialize the Plot1DController instance.

        This method initializes the Plot1DController instance by calling the constructor of the parent class (`Plot1DWidget`).
        It sets the provided data and creates the plot lines. It also sets the plot properties such as title and axis labels.

        Args:
            x_data: The x data for the plot (numpy array).
            y_data: The y data for the plot (list of numpy arrays).
            legend: The legend for each line (list of strings).
            x_label: The label for the x axis (string).
            y_label: The label for the y axis (string).
            title: The title of the plot (string).

        Returns:
            None
        """
        super(Plot1DController, self).__init__()
        self.y_data = y_data
        self.x_data = x_data
        self.x_label = x_label
        self.y_label = y_label
        self.title = title

        # Set text
        self.label2.setText("<span style='font-size: 8pt'>%s=%0.2f, %s=%0.2f</span>" % (x_label, 0, y_label, 0))

        # Add lines to plot_item
        n_lines = len(y_data)
        self.lines = []
        x_min = 0
        x_max = 0
        y_min = 0
        y_max = 0
        for line in range(n_lines):
            if type(x_data) is list:
                x = x_data[line]
            else:
                x = x_data.copy()
            y = y_data[line]
            self.lines.append(self.plot_item.plot(x, y, pen=self.pen[line], name=legend[line]))
            if line == 0:
                x_min = np.min(x)
                x_max = np.max(x)
            if np.min(x) < x_min:
                x_min = np.min(x)
            if np.max(x) > x_max:
                x_max = np.max(x)
            if np.min(y) < y_min:
                y_min = np.min(y)
            if np.max(y) > y_max:
                y_max = np.max(y)
            self.plot_item.setXRange(x_min, x_max, padding=0)
            self.plot_item.setYRange(y_min, y_max, padding=0)
            if y_min == y_max:
                self.plot_item.setYRange(-1, 1, padding=0)

        # Set the plot properties
        self.plot_item.setTitle("%s" % title)
        self.plot_item.setLabel('bottom', x_label)
        self.plot_item.setLabel('left', y_label)

    def mouseMoved(self, evt):
        """
        Handle the mouseMoved event and update the information displayed at the cursor position on the plot.

        Args:
            evt (QGraphicsSceneMouseEvent): The mouse event object.

        Returns:
            None
        """
        pos = evt[0]
        if self.plot_item.sceneBoundingRect().contains(pos):
            if type(self.x_data) is not list:
                curves = self.plot_item.listDataItems()
                x, y = curves[0].getData()
                mouse_point = self.plot_item.vb.mapSceneToView(pos)
                index = np.argmin(np.abs(self.x_data - mouse_point.x()))
                self.label2.setText("<span style='font-size: 8pt'>%s=%0.2f, %s=%0.2f</span>" % (self.x_label,
                                                                                                x[index],
                                                                                                self.y_label,
                                                                                                y[index]))
                self.crosshair_v.setPos(x[index])
                self.crosshair_h.setPos(y[index])
            else:
                mouse_point = self.plot_item.vb.mapSceneToView(pos)
                self.label2.setText("x = %0.4f, y = %0.4f" % (mouse_point.x(), mouse_point.y()))
                self.crosshair_v.setPos(mouse_point.x())
                self.crosshair_h.setPos(mouse_point.y())



from PyQt5.QtWidgets import QPushButton, QTabWidget, QWidget, QVBoxLayout, QLabel, QCheckBox, QLineEdit, QHBoxLayout, \
    QGroupBox


class PostProcessingTabWidget(QWidget):
    """
    PostProcessingTabWidget class for displaying a tab widget for post-processing options.

    Inherits from QTabWidget provided by PyQt5 to display a tab widget for post-processing options.

    Attributes:
        main: The parent widget.
    """

    def __init__(self, parent, *args, **kwargs):
        """
        Initialize the PostProcessingTabWidget.

        Args:
            parent: The parent widget.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(PostProcessingTabWidget, self).__init__(*args, **kwargs)

        # The 'main' attribute represents the parent widget, which is used to access the main window or controller.
        self.main = parent

        # Layout for BM4D tab
        self.std_label = QLabel('Standard deviation')
        self.run_filter_button = QPushButton('Run filter')
        self.auto_checkbox = QCheckBox('Auto')
        self.auto_checkbox.setChecked(True)
        self.std_text_field = QLineEdit()
        self.bm4d_layout = QVBoxLayout()
        self.std_layout = QHBoxLayout()
        self.bm4d_layout.addLayout(self.std_layout)
        self.std_layout.addWidget(self.std_label)
        self.std_layout.addWidget(self.std_text_field)
        self.bm4d_layout.addWidget(self.auto_checkbox)
        self.bm4d_layout.addWidget(self.run_filter_button)

        self.bm4d_group = QGroupBox('BM4D')
        self.bm4d_group.setLayout(self.bm4d_layout)

        # Layout for Gaussian tab
        self.gaussian_layout = QVBoxLayout()
        self.gaussian_std_layout = QHBoxLayout()
        self.gaussian_label = QLabel('Standard deviation')
        self.gaussian_text_field = QLineEdit()
        self.gaussian_std_layout.addWidget(self.gaussian_label)
        self.gaussian_std_layout.addWidget(self.gaussian_text_field)
        self.run_gaussian_button = QPushButton('Run filter')
        self.gaussian_layout.addLayout(self.gaussian_std_layout)
        self.gaussian_layout.addWidget(self.run_gaussian_button)

        self.gauss_group = QGroupBox('Gaussian filter')
        self.gauss_group.setLayout(self.gaussian_layout)

        self.post_layout = QVBoxLayout()
        self.post_layout.addWidget(self.bm4d_group)
        self.post_layout.addWidget(self.gauss_group)
        self.post_layout.addStretch()
        self.setLayout(self.post_layout)


"""
Created on Thu August 17th 2023
@author: J.M. Algarín, MRILab, i3M, CSIC, Valencia
@email: josalggui@i3m.upv.es
@Summary: code to communicate with arduino for autotuning
Specific hardware from MRILab @ i3M is required
"""

import numpy as np
import serial.tools.list_ports
import serial
from scipy.interpolate import interp1d

import time

from marge.vna import Hardware


class Arduino:
    def __init__(self, baudrate=115200, timeout=0.1, name='test', serial_number=None):
        """
        Initialize an Arduino object.

        :param baudrate: Baud rate for communication (default is 115200).
        :param timeout: Timeout for communication operations (default is 0.1 seconds).
        """
        self.device = None
        self.serial = None
        self.port = None
        self.baudrate = baudrate
        self.timeout = timeout
        self.serial_number = serial_number
        self.name = name

    def findPort(self):
        """
        Find the port of the connected Arduino.

        :return: The port of the Arduino if found, otherwise False.
        """
        arduino_port = None
        ports = serial.tools.list_ports.comports()
        for port in ports:
            if port.serial_number == self.serial_number:
                arduino_port = port.device

        if arduino_port is None:
            print("WARNING: No Arduino found for " + self.name)
            return False
        else:
            return arduino_port

    def connect(self):
        """
        Connect to the Arduino.

        :return: True if connected successfully, otherwise False.
        """
        if not self.device:
            self.port = self.findPort()
            if not self.port:
                return False
            else:
                self.device = serial.Serial(port=self.port, baudrate=self.baudrate, timeout=self.timeout)
                print("Connected to Arduino for " + self.name)
                time.sleep(1.0)

    def disconnect(self):
        """
        Disconnect from the Arduino.
        """
        if self.device is not None:
            self.device.close()
            print("Disconnected from Arduino for " + self.name)
            self.device = None

    def send(self, data):
        """
        Send data to the Arduino.

        :param data: The data to be sent.
        """
        if self.device is not None:
            self.device.write(data.encode())
        return self.receive()

    def receive(self):
        """
        Receive data from the Arduino.

        :return: The received data.
        """
        if self.device is not None:
            while self.device.in_waiting == 0:
                pass
            return self.device.readline()
        else:
            return "False".encode('utf-8')


class VNA:
    def __init__(self):
        """
        Initialize a Vectorial Network Analyzer (VNA) object.
        """
        self.connected = None
        self.frequencies = None
        self.interface = None
        self.data = []
        self.device = None

    def connect(self):
        """
        Connect to the nanoVNA device.

        :return: True if connected successfully, otherwise False.
        """
        if not self.device:
            try:
                self.interface = Hardware.get_interfaces()[0]
                self.interface.open()
                self.interface.timeout = 0.05
                time.sleep(0.1)
                self.device = Hardware.get_VNA(self.interface)
                self.frequencies = np.array(self.device.readFrequencies()) * 1e-6  # MHz
                print("Connected to nanoVNA for auto-tuning")
                return True
            except IndexError:
                print("WARNING: No interfaces available for nanoVNA auto-tuning")
                return False
            except Exception as e:
                print(f"WARNING: Failed to connect to nanoVNA for auto-tuning: {e}")
                return False
        else:
            return True

    def getFrequency(self):
        """
        Get the array of frequencies at which measurements were taken.

        :return: Array of frequencies in MHz.
        """
        if self.device is not None:
            return self.frequencies

    def getData(self):
        """
        Get the measurement data.

        :return: List of complex measurement data.
        """
        if self.device is not None:
            return self.data

    def getS11(self, f0=None):
        """
        Get S11 parameter and impedance for a specific frequency.

        :param f0: Frequency at which to get S11 parameter (in MHz).
        :return: Tuple containing S11 parameter and impedance at the given frequency.
        """
        if self.device is not None:
            self.data = []
            for value in self.device.readValues("data 0"):
                self.data.append(float(value.split(" ")[0]) + 1j * float(value.split(" ")[1]))

            # Create a linear interpolation function
            interp_func = interp1d(self.frequencies, self.data, kind='cubic')

            # Perform interpolation
            s11 = interp_func(f0)
            z11 = 50 * (1 + s11) / (1 - s11)

            return s11, z11



if __name__ == "__main__":
    device = VNA()
    device.connect()
    s11, z11 = device.getS11(2.9713)
    print(s11)
    print(z11)

    # # Create an instance of the Arduino class and connect to an Arduino
    # arduino = Arduino()
    #
    # # Disconnect from the Arduino
    # arduino.connect()
    # arduino.disconnect()



"""
@author: J.M. Algarín, february 03th 2022
MRILAB @ I3M
"""

import time

import marge.marcos.marcos_client.experiment as ex
import marge.marcos.marcos_client.server_comms as sc
import marge.configs.hw_config as hw
import numpy as np

class Experiment(ex.Experiment):
    """
    Custom experiment class that extends the base Experiment class from the 'ex' module.

    Args:
        lo_freq (float): Frequency of the LO (Local Oscillator) in MHz.
        rx_t (float): RX (Receiver) time in microseconds. Should be multiples of 1/122.88, where some values like 3.125 are exact.
        seq_dict (dict): Dictionary containing the sequence information.
        seq_csv (str): Path to a CSV file containing the sequence information.
        rx_lo (int): Specifies which internal NCO (Numerically Controlled Oscillator) local oscillator to use for each channel.
        grad_max_update_rate (float): Maximum update rate of the gradient in MSPS (Mega Samples Per Second) across all channels in parallel.
        gpa_fhdo_offset_time (int): Offset time used when GPA-FHDO (Gradient Pulse Amplitude - Fractional High Dynamic Range Output) is used.
        print_infos (bool): Flag to control the display of server info messages.
        assert_errors (bool): Flag to control whether to halt on server errors.
        init_gpa (bool): Flag to initialize the GPA (Gradient Pulse Amplitude) when the Experiment object is created.
        initial_wait (float): Initial pause before the experiment begins, in microseconds. Required to configure the LOs (Local Oscillators) and RX rate.
        auto_leds (bool): Flag to automatically scan the LED (Light-Emitting Diode) pattern from 0 to 255 as the sequence runs.
        prev_socket (socket): Previously-opened socket to maintain status.
        fix_cic_scale (bool): Flag to scale the RX (Receiver) data precisely based on the rate being used.
        set_cic_shift (bool): Flag to program the CIC (Cascaded Integrator-Comb) internal bit shift to maintain the gain within a factor of 2 independent of the rate.
        allow_user_init_cfg (bool): Flag to allow user-defined alteration of flocra (Field-Programmable Logic Controller for Real-Time Acquisition) configuration set by init.
        halt_and_reset (bool): Flag to halt any existing sequences that may be running upon connecting to the server.
        flush_old_rx (bool): Flag to read out and clear the old RX (Receiver) FIFOs before running a sequence.

    Summary:
        The Experiment class extends the base Experiment class from the 'ex' module and provides additional functionality and customization for experiments.
        It inherits all the attributes and methods from the base class and overrides the __init__() and run() methods.
    """
    def __init__(self,
                 lo_freq=1,  # MHz
                 rx_t=3.125, # us; multiples of 1/122.88, such as 3.125, are exact, others will be rounded to the nearest multiple of the 122.88 MHz clock
                 seq_dict=None,
                 seq_csv=None,
                 rx_lo=0,  # which of internal NCO local oscillators (LOs), out of 0, 1, 2, to use for each channel
                 grad_max_update_rate=0.2,  # MSPS, across all channels in parallel, best-effort
                 gpa_fhdo_offset_time=0, # when GPA-FHDO is used, offset the Y, Z and Z2 gradient times by 1x, 2x and 3x this value to emulate 'simultaneous' updates
                 print_infos=True,  # show server info messages
                 assert_errors=True,  # halt on server errors
                 init_gpa=False,  # initialise the GPA (will reset its outputs when the Experiment object is created)
                 initial_wait=None, # initial pause before experiment begins - required to configure the LOs and RX rate; must be at least a few us. Is suitably set based on grad_max_update_rate by default.
                 auto_leds=True, # automatically scan the LED pattern from 0 to 255 as the sequence runs (set to off if you wish to manually control the LEDs)
                 prev_socket=None,  # previously-opened socket, if want to maintain status etc
                 fix_cic_scale=True, # scale the RX data precisely based on the rate being used; otherwise a 2x variation possible in data amplitude based on rate
                 set_cic_shift=False, # program the CIC internal bit shift to maintain the gain within a factor of 2 independent of rate; required if the open-source CIC is used in the design
                 allow_user_init_cfg=False, # allow user-defined alteration of flocra configuration set by init, namely RX rate, LO properties etc; see the compile() method for details
                 halt_and_reset=False,  # upon connecting to the server, halt any existing sequences that may be running
                 flush_old_rx=False, # when debugging or developing new code, you may accidentally fill up the RX FIFOs - they will not automatically be cleared in case there is important data inside. Setting this true will always read them out and clear them before running a sequence. More advanced manual code can read RX from existing sequences.
                 ):
        """
        Initialize the Experiment object with the specified parameters.
        """
        super(Experiment, self).__init__(lo_freq,
                                         rx_t / hw.oversamplingFactor,
                                         seq_dict,
                                         seq_csv,
                                         rx_lo,
                                         grad_max_update_rate,
                                         gpa_fhdo_offset_time,
                                         print_infos,
                                         assert_errors,
                                         init_gpa,
                                         initial_wait,
                                         auto_leds,
                                         prev_socket,
                                         fix_cic_scale,
                                         set_cic_shift,
                                         allow_user_init_cfg,
                                         halt_and_reset,
                                         flush_old_rx,)

    def getSamplingRate(self):
        """
        Get the sampling rate of the experiment in the sequence sampling rate.

        Returns:
            float: The sampling rate in samples per second.
        """
        return self.get_rx_ts()[0] * hw.oversamplingFactor

    def get_sampling_period(self):
        """
        Get the sampling rate of the experiment in the sequence sampling rate.

        Returns:
            float: The sampling rate in samples per second.
        """
        return self.get_rx_ts()[0] * hw.oversamplingFactor  # s

    def run(self):
        """
        Compile the TX and gradient data and send everything over to the server.
        Returns the resultant data.

        Returns:
            tuple: A tuple containing the resultant data and messages.
                   The resultant data is a dictionary containing the received IQ signals for each channel in mV.
                   The messages are server messages.

        Raises:
            None
        """

        if not self._seq_compiled:
            t1 = time.time()
            self.compile()
            t2 = time.time()
            print(f"Compilation time: {round(t2-t1, 1)} s")

        if self._flush_old_rx:
            rx_data_old, _ = sc.command({'read_rx': 0}, self._s)
            # TODO: do something with RX data previously collected by the server

        rx_data, msgs = sc.command({'run_seq': self._machine_code.tobytes()}, self._s)

        rxd = rx_data[4]['run_seq']
        rxd_iq = {}

        # (1 << 24) just for the int->float conversion to be reasonable - exact value doesn't matter for now
        rx0_norm_factor = self._rx0_cic_factor / (1 << 24)
        rx1_norm_factor = self._rx0_cic_factor / (1 << 24)

        try: # Signal in millivolts with phase as it should be
            rxd_iq['rx0'] = hw.adcFactor * rx0_norm_factor * (np.array(rxd['rx0_i']).astype(np.int32).astype(float) -
                                               1j * np.array(rxd['rx0_q']).astype(np.int32).astype(float))
        except (KeyError, TypeError):
            pass

        try: # Signal in millivolts with phase as it should be
            rxd_iq['rx1'] = hw.adcFactor * rx1_norm_factor * (np.array(rxd['rx1_i']).astype(np.int32).astype(float) -
                                               1j * np.array(rxd['rx1_q']).astype(np.int32).astype(float))
        except (KeyError, TypeError):
            pass

        return rxd_iq, msgs



"""
:author:    J.M. Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain

"""


class MenuController:
    """
    Menu controller class.

    This class is responsible for controlling the menus in the application. It adds menus to the main window and
    connects actions to them.

    Methods:
        __init__(self, main): Initialize the MenuController instance.

    Attributes:
        main: The main window instance.

    """
    def __init__(self, main):
        """
        Initialize the MenuController instance.

        This method initializes the MenuController instance by setting the `main` attribute to the provided `main`
        window instance. It adds menus to the main window and connects actions to them.

        Args:
            main: The main window instance.

        Returns:
            None
        """
        self.main = main

        # Add menus
        self.menu_scanner = self.main.menu.addMenu("Scanner")
        self.menu_protocols = self.main.menu.addMenu("Protocols")
        self.menu_sequences = self.main.menu.addMenu("Sequences")
        self.menu_session = self.main.menu.addMenu("Session")

        # Protocol menu
        self.menu_protocols.addAction(self.main.toolbar_protocols.action_new_protocol)
        self.menu_protocols.addAction(self.main.toolbar_protocols.action_del_protocol)
        self.menu_protocols.addAction(self.main.toolbar_protocols.action_new_sequence)
        self.menu_protocols.addAction(self.main.toolbar_protocols.action_del_sequence)

        # Scanner menu
        self.menu_scanner.addAction(self.main.toolbar_marcos.action_start)
        self.menu_scanner.addAction(self.main.toolbar_marcos.action_copybitstream)
        self.menu_scanner.addAction(self.main.toolbar_marcos.action_server)
        self.menu_scanner.addAction(self.main.toolbar_marcos.action_gpa_init)

        # Sequences menu
        self.menu_sequences.addAction(self.main.toolbar_sequences.action_load_parameters)
        self.menu_sequences.addAction(self.main.toolbar_sequences.action_save_parameters)
        self.menu_sequences.addAction(self.main.toolbar_sequences.action_save_parameters_cal)
        self.menu_sequences.addAction(self.main.toolbar_sequences.action_add_to_list)
        self.menu_sequences.addAction(self.main.toolbar_sequences.action_acquire)
        self.menu_sequences.addAction(self.main.toolbar_sequences.action_bender)
        self.menu_sequences.addAction(self.main.toolbar_sequences.action_view_sequence)



import sys
import qdarkstyle
from PyQt5.QtCore import QThreadPool

from PyQt5.QtWidgets import QMainWindow, QWidget, QHBoxLayout, QVBoxLayout, QApplication, QStatusBar, QMenuBar, \
    QSplitter, QTextEdit, QSizePolicy

from marge.controller.controller_reconstruction import ReconstructionTabController
from marge.controller.controller_post import PostProcessingTabController
from marge.controller.controller_toolbar_figures import FiguresControllerPos
from marge.controller.controller_visualisation import VisualisationTabController
from marge.controller.controller_preprocessing import PreProcessingTabController
from marge.controller.controller_history_list import HistoryListControllerPos
from marge.controller.controller_imageview import ImageViewController
from marge.controller.controller_figures import FiguresLayoutController
from marge.controller.controller_console import ConsoleController
from marge.controller.controller_toolbar_post import ToolBarControllerPost, MainWindow_toolbar
from marge.controller.controller_processing import ProcessingController


class MainWindow(QMainWindow):
    """
    Main window class for the application.

    Inherits from QMainWindow to provide the main application window.

    Attributes:
        threadpool (QThreadPool): Thread pool for parallel running.
        main_widget (QWidget): Central widget for the main window.
        main_layout (QHBoxLayout): Main layout for the main window.
        left_layout (QVBoxLayout): Layout for the left side of the main window.
        right_layout (QVBoxLayout): Layout for the right side of the main window.
        image_view_layout (QVBoxLayout): Layout for the image view widgets.
        image_view_splitter (QSplitter): Splitter widget for the image view widgets.
        image_view_widget (ImageViewController): Image view widget for displaying images.
        history_layout (QHBoxLayout): Layout for the history list and history controller.
        history_controller (HistoryListController): Controller for the history list.
        history_widget (HistoryListWidget): Widget for displaying the history list.
        toolbar_controller (ToolBarController): Controller for the toolbar.
        tab_controller (TabController): Controller for the tab widget.
        postprocessing_controller (PostProcessingTabController): Controller for the post-processing tab.
        preprocessing_controller (PreProcessingTabController): Controller for the pre-processing tab.
        reconstruction_controller (ReconstructionTabController): Controller for the reconstruction tab.
        visualisation_controller (VisualisationTabController): Controller for the visualisation tab.
        console (ConsoleController): Controller for the console.
    """

    def __init__(self, session, main):
        """
        Initialize the MainWindow.

        Args:
            None
        """
        super().__init__()
        self.session = session
        self.main = main

        # Set stylesheet based on theme
        if self.session["black_theme"]:
            self.styleSheet = qdarkstyle.load_stylesheet_pyqt5()
        else:
            self.styleSheet = ""
        self.setStyleSheet(self.styleSheet)

        # Set window parameters
        self.setWindowTitle(session['directory'])

        # Threadpool for parallel running
        self.threadpool = QThreadPool()

        # Toolbar addition
        self.main_window = MainWindow_toolbar()
        self.toolbar_image = ToolBarControllerPost(self.main_window, self, "Image toolbar")
        self.addToolBar(self.toolbar_image)

        # Image toolbar
        self.toolbar_image_2 = FiguresControllerPos(self, "Image toolbar 2")
        self.addToolBar(self.toolbar_image_2)

        # Status bar and a menu bar adding
        self.setStatusBar(QStatusBar(self))
        self.setMenuBar(QMenuBar(self))

        # Set the central widget
        self.main_widget = QWidget()
        self.setCentralWidget(self.main_widget)

        # Set layouts
        self.main_layout = QHBoxLayout()
        self.main_widget.setLayout(self.main_layout)

        # Layout for processing and console
        self.left_layout = QVBoxLayout()
        self.main_layout.addLayout(self.left_layout)

        # Layout for figures and history list
        self.right_layout = QVBoxLayout()
        self.main_layout.addLayout(self.right_layout)

        # Tab addition
        self.tab_controller = ProcessingController(parent=self)
        self.left_layout.addWidget(self.tab_controller)

        # Image view addition
        self.image_view_widget = FiguresLayoutController(self)
        self.image_view_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.right_layout.addWidget(self.image_view_widget)

        # Layout for output history
        self.output_layout_h = QHBoxLayout()
        self.right_layout.addLayout(self.output_layout_h)

        # Add list to show the history
        self.history_list = HistoryListControllerPos(parent=self)
        self.history_list.setMaximumHeight(200)
        self.history_list.setMinimumHeight(200)
        self.output_layout_h.addWidget(self.history_list)

        # Table with list of applied methods to selected item in the history
        self.methods_list = QTextEdit()
        self.methods_list.setMaximumHeight(200)
        self.methods_list.setMinimumHeight(200)
        self.output_layout_h.addWidget(self.methods_list)



if __name__ == '__main__':
    app = QApplication(sys.argv)

    session = {'directory': '../test'}
    ui = MainWindow(session=session, main=None)
    ui.show()

    sys.exit(app.exec())



from PyQt5.QtWidgets import QPushButton, QVBoxLayout, QLabel, QLineEdit, QHBoxLayout, QGroupBox, QWidget


class ReconstructionTabWidget(QWidget):
    """
    ReconstructionTabWidget class for displaying a tab widget for image reconstruction options.

    Inherits from QTabWidget provided by PyQt5 to display a tab widget for image reconstruction options.

    Attributes:
        main: The parent widget.
    """

    def __init__(self, parent, *args, **kwargs):
        """
        Initialize the ReconstructionTabWidget.

        Args:
            parent: The parent widget.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(ReconstructionTabWidget, self).__init__(*args, **kwargs)

        # The 'main' attribute represents the parent widget, which is used to access the main window or controller.
        self.main = parent

        # ART
        self.image_art_button = QPushButton('ART')

        # Labels
        self.niter_label = QLabel('Number of iterations')
        self.lambda_label = QLabel('Lambda')

        # Text Fields
        self.niter_text_field = QLineEdit()
        self.niter_text_field.setText('1')
        self.lambda_text_field = QLineEdit()
        self.lambda_text_field.setText('1')

        # Layouts
        self.order_layout = QHBoxLayout()
        self.order_layout.addWidget(self.niter_label)
        self.order_layout.addWidget(self.niter_text_field)
        self.order_layout.addWidget(self.lambda_label)
        self.order_layout.addWidget(self.lambda_text_field)

        self.art_layout = QVBoxLayout()
        self.art_layout.addLayout(self.order_layout)
        self.art_layout.addWidget(self.image_art_button)

        self.art_group = QGroupBox('ART')
        self.art_group.setLayout(self.art_layout)

        # FFT
        self.ifft_button = QPushButton('k-space -> i-space')
        self.dfft_button = QPushButton('i-space -> k-space')

        self.fft_layout = QVBoxLayout()
        self.fft_layout.addWidget(self.ifft_button)
        self.fft_layout.addWidget(self.dfft_button)

        self.fft_group = QGroupBox('FFT')
        self.fft_group.setLayout(self.fft_layout)

        # *****************
        # POCS
        # *****************
        # Partial fourier factor
        self.partial_reconstruction_label = QLabel('Factor')
        self.partial_reconstruction_factor = QLineEdit()
        self.partial_reconstruction_factor.setPlaceholderText("Readout, Phase, Slice")
        self.partial_reconstruction_factor.setStatusTip('To apply partial acquisition use numbers between'
                                                        '0.5 and 1, otherwise use 0')

        self.factor_layout = QHBoxLayout()
        self.factor_layout.addWidget(self.partial_reconstruction_label)
        self.factor_layout.addWidget(self.partial_reconstruction_factor)

        self.nb_points_label = QLabel('Number of points')
        self.nb_points_text_field = QLineEdit()
        self.nb_points_text_field.setText('2')

        self.nb_points_layout = QHBoxLayout()
        self.nb_points_layout.addWidget(self.nb_points_label)
        self.nb_points_layout.addWidget(self.nb_points_text_field)

        self.threshold_label = QLabel('Correlation threshold')
        self.threshold_text_field = QLineEdit()
        self.threshold_text_field.setText('1e-6')

        self.threshold_layout = QHBoxLayout()
        self.threshold_layout.addWidget(self.threshold_label)
        self.threshold_layout.addWidget(self.threshold_text_field)

        self.pocs_button = QPushButton('Run POCS')
        self.zero_button = QPushButton('Run iFFT with zero padding')

        self.pocs_layout = QVBoxLayout()
        self.pocs_layout.addLayout(self.factor_layout)
        self.pocs_layout.addLayout(self.nb_points_layout)
        self.pocs_layout.addLayout(self.threshold_layout)
        self.pocs_layout.addWidget(self.pocs_button)
        self.pocs_layout.addWidget(self.zero_button)

        self.pocs_group = QGroupBox("Partial Reconstruction")
        self.pocs_group.setLayout(self.pocs_layout)

        # Main layout
        self.reconstruction_layout = QVBoxLayout()
        self.reconstruction_layout.addWidget(self.art_group)
        self.reconstruction_layout.addWidget(self.fft_group)
        self.reconstruction_layout.addWidget(self.pocs_group)
        self.reconstruction_layout.addStretch()
        self.setLayout(self.reconstruction_layout)



"""
@author:    José Miguel Algarín
@email:     josalggui@i3m.upv.es
@affiliation:MRILab, i3M, CSIC, Valencia, Spain
"""
from PyQt5.QtWidgets import QToolBar, QAction
from PyQt5.QtGui import QIcon

class SequenceToolBar(QToolBar):
    def __init__(self, parent, *args, **kwargs):
        super(SequenceToolBar, self).__init__(*args, **kwargs)
        self.main = parent

        # Autocalibration
        self.action_autocalibration = QAction(QIcon("resources/icons/calibration-light.png"), "Autocalibration", self)
        self.action_autocalibration.setStatusTip("Run autocalibration")
        self.addAction(self.action_autocalibration)

        # Localizer
        self.action_localizer = QAction(QIcon("resources/icons/localizer-light.png"), "Localizer", self)
        self.action_localizer.setStatusTip("Run Localizer")
        self.addAction(self.action_localizer)

        # Add sequence to waiting list
        self.action_add_to_list = QAction(QIcon("resources/icons/clipboard-list-check"), "Sequence to list", self)
        self.action_add_to_list.setStatusTip("Add current sequence to waiting list")
        self.addAction(self.action_add_to_list)

        # Add run action
        self.action_acquire = QAction(QIcon("resources/icons/acquire.png"), "Acquire", self)
        self.action_acquire.setStatusTip("Run current sequence")
        self.addAction(self.action_acquire)

        # Iterative run
        self.action_iterate = QAction(QIcon("resources/icons/iterate.png"), "Iterative run", self)
        self.action_iterate.setStatusTip("Set iterative mode on")
        self.addAction(self.action_iterate)

        # Bender button
        self.action_bender = QAction(QIcon("resources/icons/bender.png"), "Bender button", self)
        self.action_bender.setStatusTip("Bender will do a full protocol with a single click")
        self.addAction(self.action_bender)

        # Plot sequence
        self.action_view_sequence = QAction(QIcon("resources/icons/plotSequence.png"), "Plot sequence", self)
        self.action_view_sequence.setStatusTip("Plot current sequence")
        self.addAction(self.action_view_sequence)

        # Save sequence parameters
        self.action_save_parameters = QAction(QIcon("resources/icons/saveParameters.png"), "Save parameters", self)
        self.action_save_parameters.setStatusTip("Save the current parameters")
        self.addAction(self.action_save_parameters)

        # Load sequence parameters
        self.action_load_parameters = QAction(QIcon("resources/icons/loadParameters.png"), "Load parameters", self)
        self.action_load_parameters.setStatusTip("Load parameters and update current sequence to loaded values")
        self.addAction(self.action_load_parameters)

        # Save sequence parameters for calibration
        self.action_save_parameters_cal = QAction(QIcon("resources/icons/favouriteParameters.png"),
                                                  "Save for calibration", self)
        self.action_save_parameters_cal.setStatusTip("Save current configuration for calibration")
        self.addAction(self.action_save_parameters_cal)



"""
:author:    J.M. Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain

"""
import os
import csv
import copy

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QTableWidgetItem, QMenu, QAction

from marge.seq.sequences import defaultsequences
from marge.widgets.widget_protocol_inputs import ProtocolInputsWidget

import marge.configs.hw_config as hw

import numpy as np


class ProtocolInputsController(ProtocolInputsWidget):
    """
    Controller class for managing protocol inputs.

    Args:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Inherits:
        ProtocolInputsWidget: Base class for protocol inputs widget.
    """
    def __init__(self, *args, **kwargs):
        """
        Initializes the ProtocolInputsController.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(*args, **kwargs)

        self.clicked_item = None
        self.updateProtocolInputs()

        # Connect items to runToList method
        self.itemDoubleClicked.connect(self.sequenceDoubleClicked)
        self.itemClicked.connect(self.showSequenceInputs)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.showContextMenu)

    def showContextMenu(self, point):
        """
        Displays the context menu when right-clicked on an item.

        Args:
            point (QPoint): The position where the context menu was requested.
        """
        self.clicked_item = self.itemAt(point)
        if self.clicked_item is not None:
            menu = QMenu(self)

            # Delete action button
            action1 = QAction("Delete", self)
            action1.triggered.connect(self.deleteSequence)
            menu.addAction(action1)

            # Show input action button
            action2 = QAction("Show inputs", self)
            action2.triggered.connect(self.showProtocolInputs)
            menu.addAction(action2)

            # Load inputs action button
            action3 = QAction("Load inputs", self)
            action3.triggered.connect(self.loadProtocolInputs)
            menu.addAction(action3)

            menu.exec_(self.mapToGlobal(point))

    def deleteSequence(self):
        """
        Deletes a sequence from the protocol inputs.
        """
        protocol = self.main.protocol_list.getCurrentProtocol()
        file = "%s.csv" % self.clicked_item.text()
        path = "protocols/%s/%s" % (protocol, file)
        os.remove(path)
        self.updateProtocolInputs()
        print("Protocol removed")

    def sequenceDoubleClicked(self, item):
        """
        Handles double-click event on a sequence item.

        Args:
            item (QListWidgetItem): The item that was double-clicked.
        """
        protocol = self.main.protocol_list.getCurrentProtocol()
        protocol_item = item.text()
        file = protocol_item + ".csv"
        seq_name = protocol_item.split('_')[0]

        sequence = copy.copy(defaultsequences[seq_name])

        # Pick the shimming from sequence, as it was set up by autocalibration
        try:
            shimming = copy.copy(sequence.mapVals['shimming'])
        except:
            pass

        # Load parameters
        sequence.loadParams("protocols/"+protocol, file)

        # Set larmor frequency, fov and dfov to the value into the hw_config file
        sequence.mapVals['larmorFreq'] = hw.larmorFreq
        sequence.mapVals['fov'] = hw.fov
        sequence.mapVals['dfov'] = hw.dfov
        sequence.mapVals['shimming'] = shimming
        hw.dfov = [0.0, 0.0, 0.0]

        # Run the sequence
        map_nmspc = list(sequence.mapNmspc.values())
        map_vals = list(sequence.mapVals.values())
        self.main.toolbar_sequences.runToList(seq_name=seq_name, item_name=protocol_item, map_nmspc=map_nmspc,
                                              map_vals=map_vals)

    def updateProtocolInputs(self):
        """
        Updates the protocol inputs list.
        """
        self.clear()

        # Get predefined sequences for each protocol
        self.sequences = {}
        for protocol in self.main.protocol_list.protocols:
            # Construct the path
            path = os.path.join("protocols", protocol)

            # List all files in the specified directory
            files = [file for file in os.listdir(path) if file.endswith('.csv') and os.path.isfile(os.path.join(path, file))]

            # Sort the list of files based on their creation date
            files.sort(key=lambda x: os.path.getctime(os.path.join(path, x)))

            # Remove extensions
            files = [file.split('.')[0] for file in files]

            # Add files to the protocol variable
            self.sequences[protocol] = files.copy()

        # Add the predefined sequences of the first protocol to the protocol_input list
        protocol = self.main.protocol_list.getCurrentProtocol()
        if len(self.sequences) > 0:
            self.addItems(self.sequences[protocol])

    def showSequenceInputs(self, item):
        """
        Displays the inputs for a selected sequence.

        Args:
            item (QListWidgetItem): The selected item.
        """
        # Get file name
        file_name = "%s.csv" % item.text()
        seq_name = file_name.split('_')[0]

        # Extract items from the input_data
        input_info = defaultsequences[seq_name].mapNmspc.values()
        input_vals = defaultsequences[seq_name].mapVals.values()

        # Set number of rows
        self.main.input_table.setColumnCount(1)
        self.main.input_table.setRowCount(len(input_info))

        # Input items into the table
        self.main.input_table.setVerticalHeaderLabels(input_info)
        self.main.input_table.setHorizontalHeaderLabels(['Values'])
        for m, item in enumerate(input_vals):
            new_item = QTableWidgetItem(str(item))
            self.main.input_table.setItem(m, 0, new_item)

    def showProtocolInputs(self):

        # Get path to the file
        protocol = self.main.protocol_list.getCurrentProtocol()
        file_name = "%s.csv" % self.clicked_item.text()
        path_to_file = "protocols/%s/%s" % (protocol, file_name)

        # Get strings for the info
        seq_name = file_name.split('_')[0]
        input_info = defaultsequences[seq_name].mapNmspc.values()

        # Get value for the info
        map_vals = {}
        with open(path_to_file, 'r') as csvfile:
            reader = csv.DictReader(csvfile)
            for l in reader:
                map_vals = l
        input_vals = map_vals.values()

        # Set number of rows
        self.main.input_table.setColumnCount(1)
        self.main.input_table.setRowCount(len(input_info))

        # Input items into the table
        self.main.input_table.setVerticalHeaderLabels(input_info)
        self.main.input_table.setHorizontalHeaderLabels(['Values'])
        for m, item in enumerate(input_vals):
            new_item = QTableWidgetItem(str(item))
            self.main.input_table.setItem(m, 0, new_item)

    def loadProtocolInputs(self):
        # Get path to the file
        protocol = self.main.protocol_list.getCurrentProtocol()
        file_name = "%s.csv" % self.clicked_item.text()
        path_to_file = "protocols/%s/%s" % (protocol, file_name)

        # Get strings for the info
        seq_name = file_name.split('_')[0]
        seq = defaultsequences[seq_name]
        map_vals_old = seq.mapVals

        # Get value for the info
        map_vals_new = {}
        with open(path_to_file, 'r') as csvfile:
            reader = csv.DictReader(csvfile)
            for l in reader:
                map_vals_new = l

        seq.mapVals = {}

        # Get key for corresponding modified parameter
        for key in seq.mapKeys:
            data_len = seq.mapLen[key]
            val_old = map_vals_old[key]
            try:
                val_new = map_vals_new[key]
            except:
                val_new = str(val_old)*1
            val_new = val_new.replace('[', '')
            val_new = val_new.replace(']', '')
            val_new = val_new.split(',')
            if type(val_old) == str:
                val_old = [val_old]
            elif data_len == 1:
                val_old = [val_old]
            data_type = type(val_old[0])

            inputNum = []
            for ii in range(data_len):
                if data_type == float or data_type == np.float64:
                    try:
                        inputNum.append(float(val_new[ii]))
                    except:
                        inputNum.append(float(val_old[ii]))
                elif data_type == int:
                    try:
                        inputNum.append(int(val_new[ii]))
                    except:
                        inputNum.append(int(val_old[ii]))
                else:
                    try:
                        inputNum.append(str(val_new[0]))
                        break
                    except:
                        inputNum.append(str(val_old[0]))
                        break
            if data_type == str:
                seq.mapVals[key] = inputNum[0]
            else:
                if data_len == 1:  # Save value into mapVals
                    seq.mapVals[key] = inputNum[0]
                else:
                    seq.mapVals[key] = inputNum

        self.main.sequence_list.updateSequence()
        print("Parameters of %s sequence loaded" % file_name)



#  NanoVNASaver
#  A python program to view and export Touchstone data from a NanoVNA
#  Copyright (C) 2019, 2020  Rune B. Broberg
#  Copyright (C) 2020,2021 NanoVNA-Saver Authors
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
import re

logger = logging.getLogger(__name__)


class Version:
    RXP = re.compile(r"""^
        \D*
        (?P<major>\d+)\.
        (?P<minor>\d+)\.?
        (?P<revision>\d+)?
        (?P<note>.*)
        $""", re.VERBOSE)

    def __init__(self, vstring: str = "0.0.0"):
        self.data = {
            "major": 0,
            "minor": 0,
            "revision": 0,
            "note": "",
        }
        try:
            self.data = Version.RXP.search(vstring).groupdict()
            for name in ("major", "minor", "revision"):
                self.data[name] = int(self.data[name])
        except TypeError:
            self.data["revision"] = 0
        except AttributeError:
            logger.error("Unable to parse version: %s", vstring)

    def __gt__(self, other: "Version") -> bool:
        l, r = self.data, other.data
        for name in ("major", "minor", "revision"):
            if l[name] > r[name]:
                return True
            if l[name] < r[name]:
                return False
        return False

    def __lt__(self, other: "Version") -> bool:
        return other.__gt__(self)

    def __ge__(self, other: "Version") -> bool:
        return self.__gt__(other) or self.__eq__(other)

    def __le__(self, other: "Version") -> bool:
        return other.__gt__(self) or self.__eq__(other)

    def __eq__(self, other: "Version") -> bool:
        return self.data == other.data

    def __str__(self) -> str:
        return (f'{self.data["major"]}.{self.data["minor"]}'
                f'.{self.data["revision"]}{self.data["note"]}')

    @property
    def major(self) -> int:
        return self.data["major"]

    @property
    def minor(self) -> int:
        return self.data["minor"]

    @property
    def revision(self) -> int:
        return self.data["revision"]

    @property
    def note(self) -> str:
        return self.data["note"]



import sys
import csv
from marge.configs import hw_config as hw
from PyQt5.QtWidgets import (
    QApplication, QWidget, QVBoxLayout, QLineEdit, QPushButton,
    QHBoxLayout, QLabel, QGridLayout
)

class OthersWidget(QWidget):
    def __init__(self):
        super().__init__()

        # Main layout
        self.main_layout = QVBoxLayout()
        self.layout = QGridLayout()
        self.main_layout.addLayout(self.layout)

        # Parameters to save inputs
        self.labels = []
        self.values = []
        self.tips = []

        # Add inputs
        self.add_input(label="Scanner name", value="Demo", tip="Name of the MRI scanner")
        self.add_input(label="FOVx (cm)", value="20.0", tip="Field of View in the X direction")
        self.add_input(label="FOVy (cm)", value="20.0", tip="Field of View in the Y direction")
        self.add_input(label="FOVz (cm)", value="20.0", tip="Field of View in the Z direction")
        self.add_input(label="Shimming factor", value="1e-5", tip="Factor used for shimming adjustments")
        self.add_input(label="Bash path", value="gnome-terminal", tip="Path for executing bash commands")
        self.add_input(label="Arduino autotuning", value="242353133363518050E0",
                       tip="Arduino serial number for autotuning")
        self.add_input(label="Arduino interlock", value="242353133363518050E1",
                       tip="Arduino serial number for interlock system")
        self.add_input(label="Arduino attenuator", value="242353133363518050E2",
                       tip="Arduino serial number for RF attenuation")

        # Dictionary to store references to input fields
        self.input_boxes = {}

        # Create blocks iteratively
        for row, (label, value) in enumerate(zip(self.labels, self.values)):
            label_widget = QLabel(label)
            input_box = QLineEdit(value)
            input_box.setStatusTip(self.tips[row])
            self.input_boxes[label] = input_box

            self.layout.addWidget(label_widget, row, 0)  # Label in column 0
            self.layout.addWidget(input_box, row, 1)  # Input box in column 1

        # Buttons
        self.save_button = QPushButton('Save', self)
        self.save_button.clicked.connect(self.save_others_entries)

        layout = QHBoxLayout()
        layout.addWidget(self.save_button)
        self.main_layout.addLayout(layout)
        self.main_layout.addStretch()

        self.setLayout(self.main_layout)
        self.setWindowTitle('Others Entry')
        self.resize(400, 200)

        # Load saved gradient entries
        self.load_others_entries()

        # Update hardware
        self.update_hw_config_others()

    def add_input(self, label="", value="", tip=""):
        self.labels.append(label)
        self.values.append(value)
        self.tips.append(tip)

    def update_hw_config_others(self):
        hw.fov = [float(self.input_boxes["FOVx (cm)"].text()),
                  float(self.input_boxes["FOVy (cm)"].text()),
                  float(self.input_boxes["FOVz (cm)"].text())]
        hw.shimming_factor = float(self.input_boxes["Shimming factor"].text())
        hw.scanner_name = self.input_boxes["Scanner name"].text()
        hw.bash_path = self.input_boxes["Bash path"].text()
        hw.ard_sn_interlock = self.input_boxes["Arduino interlock"].text()
        hw.ard_sn_attenuator = self.input_boxes["Arduino attenuator"].text()
        hw.ard_sn_autotuning = self.input_boxes["Arduino autotuning"].text()

    def save_others_entries(self):
        file_name = "configs/hw_others.csv"
        with open(file_name, 'w', newline='') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(["ID", "Value"])
            for label, input_box in self.input_boxes.items():
                writer.writerow([label, input_box.text()])  # Write each pair
        self.update_hw_config_others()
        print(f"Data saved for others.")

    def load_others_entries(self):
        """Load label-value pairs from a CSV file and update the input fields."""
        file_path = "configs/hw_others.csv"
        try:
            with open(file_path, mode="r", newline="") as file:
                reader = csv.reader(file)
                next(reader)  # Skip header row

                for row in reader:
                    if len(row) == 2:  # Ensure row has two columns
                        label, value = row
                        if label in self.input_boxes:
                            self.input_boxes[label].setText(value)  # Update input box
        except:
            print("No hardware configuration loaded for others hardware.")


if __name__ == '__main__':
    # Run the application
    app = QApplication(sys.argv)
    widget = OthersWidget()
    widget.show()
    sys.exit(app.exec())


"""
@author:    José Miguel Algarín
@email:     josalggui@i3m.upv.es
@affiliation:MRILab, i3M, CSIC, Valencia, Spain
"""
from PyQt5.QtWidgets import QComboBox, QSizePolicy
from marge.seq.sequences import defaultsequences


class SequenceListWidget(QComboBox):
    def __init__(self, parent, *args, **kwargs):
        super(SequenceListWidget, self).__init__(*args, **kwargs)
        self.main = parent

        # Add sequences to sequences list
        self.addItems(sorted(list(defaultsequences.keys())))

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self.setMaximumWidth(400)



#  NanoVNASaver
#
#  A python program to view and export Touchstone data from a NanoVNA
#  Copyright (C) 2019, 2020  Rune B. Broberg
#  Copyright (C) 2020,2021 NanoVNA-Saver Authors
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
import platform
from struct import pack, unpack_from
from time import sleep
from typing import List

from marge.vna.Serial import Interface
from marge.vna.VNA import VNA
from marge.vna.Version import Version

if platform.system() != 'Windows':
    import tty

logger = logging.getLogger(__name__)

_CMD_NOP = 0x00
_CMD_INDICATE = 0x0d
_CMD_READ = 0x10
_CMD_READ2 = 0x11
_CMD_READ4 = 0x12
_CMD_READFIFO = 0x18
_CMD_WRITE = 0x20
_CMD_WRITE2 = 0x21
_CMD_WRITE4 = 0x22
_CMD_WRITE8 = 0x23
_CMD_WRITEFIFO = 0x28

_ADDR_SWEEP_START = 0x00
_ADDR_SWEEP_STEP = 0x10
_ADDR_SWEEP_POINTS = 0x20
_ADDR_SWEEP_VALS_PER_FREQ = 0x22
_ADDR_RAW_SAMPLES_MODE = 0x26
_ADDR_VALUES_FIFO = 0x30
_ADDR_DEVICE_VARIANT = 0xf0
_ADDR_PROTOCOL_VERSION = 0xf1
_ADDR_HARDWARE_REVISION = 0xf2
_ADDR_FW_MAJOR = 0xf3
_ADDR_FW_MINOR = 0xf4

WRITE_SLEEP = 0.05

_ADF4350_TXPOWER_DESC_MAP = {
    0: '9dB attenuation',
    1: '6dB attenuation',
    2: '3dB attenuation',
    3: 'Maximum',
}
_ADF4350_TXPOWER_DESC_REV_MAP = {
    value: key for key, value in _ADF4350_TXPOWER_DESC_MAP.items()}


class NanoVNA_V2(VNA):
    name = "NanoVNA-V2"
    valid_datapoints = (101, 11, 51, 201, 301, 501, 1023)
    screenwidth = 320
    screenheight = 240

    def __init__(self, iface: Interface):
        super().__init__(iface)

        if platform.system() != 'Windows':
            tty.setraw(self.serial.fd)

        # reset protocol to known state
        with self.serial.lock:
            self.serial.write(pack("<Q", 0))
            sleep(WRITE_SLEEP)

        # firmware major version of 0xff indicates dfu mode
        if self.version.data["major"] == 0xff:
            raise IOError('Device is in DFU mode')

        if "S21 hack" in self.features:
            self.valid_datapoints = (101, 11, 51, 201, 301, 501, 1021)

        self.sweepStartHz = 200e6
        self.sweepStepHz = 1e6

        self._sweepdata = []
        self._updateSweep()

    def getCalibration(self) -> str:
        return "Unknown"

    def read_features(self):
        self.features.add("Customizable data points")
        # TODO: more than one dp per freq
        self.features.add("Multi data points")
        self.board_revision = self.read_board_revision()
        if self.board_revision >= Version("2.0.4"):
            self.sweep_max_freq_Hz = 4400e6
        else:
            self.sweep_max_freq_Hz = 3000e6
        if self.version <= Version("1.0.1"):
            logger.debug("Hack for s21 oddity in first sweeppoint")
            self.features.add("S21 hack")
        if self.version >= Version("1.0.2"):
            self.features.update({"Set TX power partial", "Set Average"})
            # Can only set ADF4350 power, i.e. for >= 140MHz
            self.txPowerRanges = [
                ((140e6, self.sweep_max_freq_Hz),
                 [_ADF4350_TXPOWER_DESC_MAP[value] for value in (3, 2, 1, 0)]),
            ]

    def readFirmware(self) -> str:
        result = f"HW: {self.read_board_revision()}\nFW: {self.version}"
        logger.debug("readFirmware: %s", result)
        return result

    def readFrequencies(self) -> List[int]:
        return [
            int(self.sweepStartHz + i * self.sweepStepHz)
            for i in range(self.datapoints)
        ]

    def _read_pointstoread(self, pointstoread, arr) -> None:
        freq_index = -1

        for i in range(pointstoread):
            (fwd_real, fwd_imag, rev0_real, rev0_imag, rev1_real,
             rev1_imag, freq_index) = unpack_from(
                "<iiiiiihxxxxxx", arr, i * 32)
            fwd = complex(fwd_real, fwd_imag)
            refl = complex(rev0_real, rev0_imag)
            thru = complex(rev1_real, rev1_imag)
            if i == 0:
                logger.debug("Freq index from: %i", freq_index)
            self._sweepdata[freq_index] = (refl / fwd, thru / fwd)

        logger.debug("Freq index to: %i", freq_index)

    def readValues(self, value) -> List[str]:
        # Actually grab the data only when requesting channel 0.
        # The hardware will return all channels which we will store.
        if value == "data 0":
            s21hack = 1 if "S21 hack" in self.features else 0
            # reset protocol to known state
            timeout = self.serial.timeout
            with self.serial.lock:
                self.serial.write(pack("<Q", 0))
                sleep(WRITE_SLEEP)
                # cmd: write register 0x30 to clear FIFO
                self.serial.write(pack("<BBB",
                                       _CMD_WRITE, _ADDR_VALUES_FIFO, 0))
                sleep(WRITE_SLEEP)
                # clear sweepdata
                self._sweepdata = [(complex(), complex())] * (
                    self.datapoints + s21hack)
                pointstodo = self.datapoints + s21hack
                # we read at most 255 values at a time and the time required
                # empirically is just over 3 seconds for 101 points or
                # 7 seconds for 255 points
                self.serial.timeout = min(pointstodo, 255) * 0.035 + 0.1
                while pointstodo > 0:
                    logger.info("reading values")
                    pointstoread = min(255, pointstodo)
                    # cmd: read FIFO, addr 0x30
                    self.serial.write(
                        pack("<BBB",
                             _CMD_READFIFO, _ADDR_VALUES_FIFO,
                             pointstoread))
                    sleep(WRITE_SLEEP)
                    # each value is 32 bytes
                    nBytes = pointstoread * 32

                    # serial .read() will try to read nBytes bytes in
                    # timeout secs
                    arr = self.serial.read(nBytes)
                    if nBytes != len(arr):
                        logger.warning("expected %d bytes, got %d",
                                       nBytes, len(arr))
                        # the way to retry on timeout is keep the data
                        # already read then try to read the rest of
                        # the data into the array
                        if nBytes > len(arr):
                            arr = arr + self.serial.read(nBytes - len(arr))
                    if nBytes != len(arr):
                        return []

                    self._read_pointstoread(pointstoread, arr)

                    pointstodo = pointstodo - pointstoread
            self.serial.timeout = timeout

            if s21hack:
                self._sweepdata = self._sweepdata[1:]

        idx = 1 if value == "data 1" else 0
        return [
            f'{str(x[idx].real)} {str(x[idx].imag)}'
            for x in self._sweepdata
        ]

    def resetSweep(self, start: int, stop: int):
        self.setSweep(start, stop)

    def _read_version(self, cmd_0: int, cmd_1: int):
        cmd = pack("<BBBB", _CMD_READ, cmd_0, _CMD_READ, cmd_1)
        with self.serial.lock:
            self.serial.write(cmd)
            sleep(WRITE_SLEEP)
            resp = self.serial.read(2)
        if len(resp) != 2:
            logger.error("Timeout reading version registers. Got: %s", resp)
            raise IOError("Timeout reading version registers")
        return Version(f"{resp[0]}.0.{resp[1]}")

    def readVersion(self) -> 'Version':
        result = self._read_version(_ADDR_FW_MAJOR,
                                    _ADDR_FW_MINOR)
        logger.debug("readVersion: %s", result)
        return result

    def read_board_revision(self) -> 'Version':
        result = self._read_version(_ADDR_DEVICE_VARIANT,
                                    _ADDR_HARDWARE_REVISION)
        logger.debug("read_board_revision: %s", result)
        return result

    def setSweep(self, start, stop):
        step = (stop - start) / (self.datapoints - 1)
        if start == self.sweepStartHz and step == self.sweepStepHz:
            return
        self.sweepStartHz = start
        self.sweepStepHz = step
        logger.info('NanoVNAV2: set sweep start %d step %d',
                    self.sweepStartHz, self.sweepStepHz)
        self._updateSweep()
        return

    def _updateSweep(self):
        s21hack = "S21 hack" in self.features
        cmd = pack("<BBQ", _CMD_WRITE8, _ADDR_SWEEP_START,
                   max(50000,
                       int(self.sweepStartHz - (self.sweepStepHz * s21hack))))
        cmd += pack("<BBQ", _CMD_WRITE8,
                    _ADDR_SWEEP_STEP, int(self.sweepStepHz))
        cmd += pack("<BBH", _CMD_WRITE2,
                    _ADDR_SWEEP_POINTS, self.datapoints + s21hack)
        cmd += pack("<BBH", _CMD_WRITE2,
                    _ADDR_SWEEP_VALS_PER_FREQ, 1)
        with self.serial.lock:
            self.serial.write(cmd)
            sleep(WRITE_SLEEP)

    def setTXPower(self, freq_range, power_desc):
        if freq_range[0] != 140e6:
            raise ValueError('Invalid TX power frequency range')
        # 140MHz..max => ADF4350
        self._set_register(0x42, _ADF4350_TXPOWER_DESC_REV_MAP[power_desc], 1)

    def _set_register(self, addr, value, size):
        packet = b''
        if size == 1:
            packet = pack("<BBB", _CMD_WRITE, addr, value)
        elif size == 2:
            packet = pack("<BBH", _CMD_WRITE2, addr, value)
        elif size == 4:
            packet = pack("<BBI", _CMD_WRITE4, addr, value)
        elif size == 8:
            packet = pack("<BBQ", _CMD_WRITE8, addr, value)
        self.serial.write(packet)
        logger.debug("set register %02x (size %d) to %x", addr, size, value)



"""
:author:    J.M. Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain

"""
import csv
import shutil
import os
import sys
import subprocess

import qdarkstyle

import marge.configs.hw_config as hw
from .controller_main import MainController
from marge.ui.window_session import SessionWindow


from marge.controller.controller_console import ConsoleController



class SessionController(SessionWindow):
    """
    Controller class for managing the session.

    Inherits:
        SessionWindow: Base class for the session window.
    """
    def __init__(self):
        """
        Initializes the SessionController.
        """
        super().__init__()
        self.console = ConsoleController()  # Initialisation console si nécessaire

        self.session = None
        self.main_gui = None
        self.tab_session.rf_coil_combo_box.addItems(hw.antenna_dict.keys())

        # Set slots for toolbar actions
        self.launch_gui_action.triggered.connect(self.runMainGui)
        self.demo_gui_action.triggered.connect(self.runDemoGui)
        self.update_action.triggered.connect(self.update_hardware)
        self.close_action.triggered.connect(self.close)
        self.switch_theme_action.triggered.connect(self.switch_theme)


        # Check if system is ready
        self.check_system()

    def check_system(self):
        check = True
        if not os.path.exists("configs/sys_projects.csv"):
            print("ERROR: Projects not configured. Add at least one project.")
            check = False
        else:
            self.check_projects.setChecked(True)
        if not os.path.exists("configs/sys_study.csv"):
            print("ERROR: Study cases not configured. Add at least one study case.")
            check = False
        else:
            self.check_study.setChecked(True)
        if not os.path.exists("configs/hw_gradients.csv"):
            print("ERROR: Gradient hardware not configured. Go to hardware config.")
            check = False
        else:
            self.check_gradients.setChecked(True)
        if not os.path.exists("configs/hw_others.csv"):
            print("ERROR: Other hardware not configured. Go to hardware config.")
            check = False
        else:
            self.check_others.setChecked(True)
        if not os.path.exists("configs/hw_redpitayas.csv"):
            print("ERROR: Red Pitayas not configured. Go to hardware config.")
            check = False
        else:
            self.check_redpitayas.setChecked(True)
        if not os.path.exists("configs/hw_rf.csv"):
            print("ERROR: RF hardware not configured. Go to hardware config.")
            check = False
        else:
            self.check_rf.setChecked(True)
        if len(hw.rp_ip_list) == 0:
            print("ERROR: Red pitaya ip address required. Go to hardware config.")
            check = False
        else:
            self.check_rp_ips.setChecked(True)
        if len(hw.antenna_dict) == 0:
            print("ERROR: Antenna definition required. Go to hardware config.")
            check = False
        else:
            self.check_rf_coils.setChecked(True)

        if check:
            print("READY: System configuration checks succeeded.")
            self.launch_gui_action.setDisabled(False)
            self.demo_gui_action.setDisabled(False)

    def update_hardware(self):
        self.tab_session.rf_coil_combo_box.clear()
        self.tab_session.rf_coil_combo_box.addItems(hw.antenna_dict.keys())
        self.check_system()

    def runMainGui(self):
        """
        Runs the main GUI and sets up the session.
        """
        self.updateSessionDict()

        # Create folder
        self.session['directory'] = os.path.join(
            'experiments', 'acquisitions',
            self.session['project'], self.session['subject_id'], self.session['study'], self.session['side'])
        if not os.path.exists(self.session['directory']):
            os.makedirs(self.session['directory'])

        # Save session in csv and copy config files
        with open(os.path.join(self.session['directory'], "session.csv"), mode="w", newline="") as file:
            writer = csv.writer(file)
            for key, value in self.session.items():
                writer.writerow([key, value])

        # Copy configuration files to session directory
        shutil.copy2("configs/hw_gradients.csv", os.path.join(self.session["directory"], "hw_gradients.csv"))
        shutil.copy2("configs/hw_others.csv", os.path.join(self.session["directory"], "hw_others.csv"))
        shutil.copy2("configs/hw_redpitayas.csv", os.path.join(self.session["directory"], "hw_redpitayas.csv"))
        shutil.copy2("configs/hw_rf.csv", os.path.join(self.session["directory"], "hw_rf.csv"))
        shutil.copy2("configs/sys_projects.csv", os.path.join(self.session["directory"], "sys_projects.csv"))
        shutil.copy2("configs/sys_study.csv", os.path.join(self.session["directory"], "sys_study.csv"))

        self.session['seriesNumber'] = 0

        # Open the main gui
        if self.main_gui is None:
            self.main_gui = MainController(session=self.session, demo=False, parent=self)
        else:
            self.main_gui.set_session(self.session)
            self.main_gui.history_list.delete_items()
            self.main_gui.console.clear_console()
            self.main_gui.set_demo(False)

        self.hide()
        self.main_gui.show()

    def runDemoGui(self):
        """
        Runs the main GUI in DEMO mode and sets up the session.
        """
        self.updateSessionDict()

        # Create folder
        self.session['directory'] = os.path.join(
            'experiments', 'acquisitions',
            self.session['project'], self.session['subject_id'], self.session['study'], self.session['side'])
        if not os.path.exists(self.session['directory']):
            os.makedirs(self.session['directory'])

        # Save session in csv and copy config files
        with open(os.path.join(self.session['directory'], "session.csv"), mode="w", newline="") as file:
            writer = csv.writer(file)
            for key, value in self.session.items():
                writer.writerow([key, value])

        shutil.copy2("configs/hw_gradients.csv", os.path.join(self.session["directory"], "hw_gradients.csv"))
        shutil.copy2("configs/hw_others.csv", os.path.join(self.session["directory"], "hw_others.csv"))
        shutil.copy2("configs/hw_redpitayas.csv", os.path.join(self.session["directory"], "hw_redpitayas.csv"))
        shutil.copy2("configs/hw_rf.csv", os.path.join(self.session["directory"], "hw_rf.csv"))
        shutil.copy2("configs/sys_projects.csv", os.path.join(self.session["directory"], "sys_projects.csv"))
        shutil.copy2("configs/sys_study.csv", os.path.join(self.session["directory"], "sys_study.csv"))

        self.session['seriesNumber'] = 0

        # Open the main gui
        if self.main_gui is None:
            self.main_gui = MainController(session=self.session, demo=True, parent=self)
        else:
            self.main_gui.set_session(self.session)
            self.main_gui.history_list.delete_items()
            self.main_gui.console.clear_console()
            self.main_gui.set_demo(True)

        self.hide()
        self.main_gui.show()

    def closeEvent(self, event):
        """
        Handle the window close event cleanly.

        Args:
            event: The close event.
        """
        if self.main_gui is not None:
            self.main_gui.app_open = False
            if not self.main_gui.demo:
                # Close server
                try:
                    subprocess.run([hw.bash_path, "--", "./communicateRP.sh", hw.rp_ip_address, "killall marcos_server"])
                except Exception as e:
                    print("ERROR: Server connection not found! Please verify if the blue LED is illuminated on the Red Pitaya.")
                    print(str(e))

                # Disable power modules
                try:
                    self.main_gui.toolbar_marcos.arduino.send("GPA_ON 0;")
                    self.main_gui.toolbar_marcos.arduino.send("RFPA_RF 0;")
                except Exception as e:
                    print("ERROR: Could not disable power modules.")
                    print(str(e))

        # Close console logging if exists
        if hasattr(self, 'console'):
            self.console.close_log()

        print('GUI closed successfully!')
        super().closeEvent(event)

    def close(self):
        """
        Closes the session and exits the program.
        """
        if self.main_gui is not None:
            self.main_gui.app_open = False
            if not self.main_gui.demo:
                # Close server
                try:
                    subprocess.run(
                        [hw.bash_path, "--", "./communicateRP.sh", hw.rp_ip_address, "killall marcos_server"])
                except Exception as e:
                    print("ERROR: Server connection not found! Please verify if the blue LED is illuminated on the Red Pitaya.")
                    print(str(e))

                # Disable power modules
                try:
                    self.main_gui.toolbar_marcos.arduino.send("GPA_ON 0;")
                    self.main_gui.toolbar_marcos.arduino.send("RFPA_RF 0;")
                except Exception as e:
                    print("ERROR: Could not disable power modules.")
                    print(str(e))

        # Close console logging if exists
        if hasattr(self, 'console'):
            self.console.close_log()

        print('GUI closed successfully!')
        sys.exit()

    def switch_theme(self):
        self.is_dark_theme = not self.is_dark_theme
        if self.is_dark_theme:
            self.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
        else:
            self.setStyleSheet("")  # Light theme: default Qt style

    def updateSessionDict(self):
        """
        Updates the session dictionary with the current session information.
        """

        def get_text_or_placeholder(widget):
            return widget.text() if widget.text() else widget.placeholderText()

        self.session = {'project': self.tab_session.project_combo_box.currentText(),
                        'study': self.tab_session.study_combo_box.currentText(),
                        'side': self.tab_session.side_combo_box.currentText(),
                        'orientation': self.tab_session.orientation_combo_box.currentText(),
                        'subject_id': get_text_or_placeholder(self.tab_session.id_line_edit),
                        'study_id': get_text_or_placeholder(self.tab_session.idS_line_edit),
                        'subject_name': get_text_or_placeholder(self.tab_session.name_line_edit),
                        'subject_surname': get_text_or_placeholder(self.tab_session.surname_line_edit),
                        'subject_birthday': get_text_or_placeholder(self.tab_session.birthday_line_edit),
                        'subject_weight': get_text_or_placeholder(self.tab_session.weight_line_edit),
                        'subject_height': get_text_or_placeholder(self.tab_session.height_line_edit),
                        'scanner': get_text_or_placeholder(self.tab_session.scanner_line_edit),
                        'rf_coil': self.tab_session.rf_coil_combo_box.currentText(),
                        'software_version': get_text_or_placeholder(self.tab_session.software_line_edit),
                        'black_theme': self.is_dark_theme}

        # Save session theme

        hw.b1Efficiency = hw.antenna_dict.get(self.session['rf_coil'], 1.0)



"""
:author:    J.M. Algarín
:email:     josalggui@i3m.upv.es
:affiliation: MRILab, i3M, CSIC, Valencia, Spain

"""
import os

from marge.widgets.widget_protocol_list import ProtocolListWidget


class ProtocolListController(ProtocolListWidget):
    """
    Controller class for managing the protocol list.

    Args:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Inherits:
        ProtocolListWidget: Base class for protocol list widget.
    """
    def __init__(self, *args, **kwargs):
        """
        Initializes the ProtocolListController.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(ProtocolListController, self).__init__(*args, **kwargs)
        self.protocol = None
        self.protocols = None

        if not os.path.exists('protocols'):
            os.makedirs('protocols')

        self.updateProtocolList()

        self.currentTextChanged.connect(self.updateProtocolInputs)

    def getCurrentProtocol(self):
        """
        Returns the currently selected protocol.

        Returns:
            str: The name of the current protocol.
        """
        return self.currentText()

    def updateProtocolInputs(self):
        """
        Updates the protocol inputs based on the selected protocol.
        """
        # Get the name of the selected sequence
        self.protocol = self.getCurrentProtocol()

        # Delete sequences from current protocol
        self.main.protocol_inputs.clear()

        # Add items corresponding to selected protocol
        self.main.protocol_inputs.addItems(self.main.protocol_inputs.sequences[self.protocol])

        return 0

    def updateProtocolList(self):
        """
        Updates the list of protocols.
        """
        self.blockSignals(True)
        self.clear()

        # Get the protocols
        self.protocols = []
        for path in os.listdir("protocols"):
            if len(path.split('.')) == 1:
                self.protocols.append(path)

        # Add protocols to list
        self.addItems(self.protocols)
        self.blockSignals(False)

        self.protocol = self.getCurrentProtocol()


import numpy as np
from pydicom import dcmread
from pydicom.data import get_testdata_file
import datetime


class DICOMImage:
    def __init__(self, path=None):
        # Cargar el archivo de prueba de DICOM
        self.meta_data = {}
        if path is None:
            path = get_testdata_file("MR_small.dcm")
        try:
            self.ds = dcmread(path)
        except:
            self.ds = dcmread(get_testdata_file("MR_small.dcm"))

    def image2Dicom(self):
        for key in self.meta_data.keys():
            setattr(self.ds, key, self.meta_data[key])

    def save(self, filename):
        self.ds.save_as(filename)

    def __str__(self):
        return f"Imagen DICOM con nombre del paciente {self.ds.PatientName} y ID {self.ds.PatientID}"


if __name__ == '__main__':
    #### Random image
    name = 'Random'
    nRd = 90
    nPh = 90
    allSlices = 5
    arr = np.zeros((allSlices, nPh, nRd), dtype=np.int16)
    arr[0, 0:20, 0:20] = 50
    arr[4, 0:40, 0:40] = 100
    repetitionTime = 200
    echoTime = 20
    ETL = 5

    #### Crear una nueva imagen DICOM
    # Image data
    dicom_image = DICOMImage()
    dicom_image.meta_data["PixelData"] = arr.tobytes()

    # Date and time
    current_time = datetime.datetime.now()
    dicom_image.meta_data["StudyDate"] = current_time.strftime("%Y%m%d")
    dicom_image.meta_data["StudyTime"] = current_time.strftime("%H%M%S")

    # Sequence parameters
    dicom_image.meta_data["Columns"] = nRd
    dicom_image.meta_data["Rows"] = nPh
    dicom_image.meta_data["NumberOfSlices"] = allSlices
    dicom_image.meta_data["RepetitionTime"] = repetitionTime
    dicom_image.meta_data["EchoTime"] = echoTime
    dicom_image.meta_data["EchoTrainLength"] = ETL

    # More DICOM tags
    # dicom_image.meta_data["SeriesNumber"] = 1
    dicom_image.meta_data["PatientName"] = " "
    dicom_image.meta_data["PatientID"] = name
    dicom_image.meta_data["PatientSex"] = " "
    dicom_image.meta_data["StudyID"] = "KneeProtocols"
    dicom_image.meta_data["InstitutionName"] = "PhysioMRI"
    dicom_image.meta_data["ImageComments"] = " "
    SOPInstanceUID = name
    dicom_image.meta_data["SOPInstanceUID"] = SOPInstanceUID

    #### Create DICOM file
    #### DICOM 3.0 ####
    dicom_image.meta_data["NumberOfFrames"] = allSlices
    dicom_image.image2Dicom()
    nameDcmFile = name + ".dcm"
    dicom_image.save("C:/Users/Physio MRI/Desktop/Dicom prueva/Ejemplo_1.dcm")



from marge.widgets.widget_processing import ProcessingWidget


class ProcessingController(ProcessingWidget):
    """
    Controller class for the TabWidget.

    Inherits from TabWidget to provide additional functionality for managing tabs.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the TabController.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super(ProcessingController, self).__init__(*args, **kwargs)



"""
@author:    José Miguel Algarín
@email:     josalggui@i3m.upv.es
@affiliation:MRILab, i3M, CSIC, Valencia, Spain
"""
import os

from PyQt5.QtWidgets import QListWidget


class ProtocolInputsWidget(QListWidget):
    def __init__(self, main):
        super().__init__()
        self.main = main

